import bpy
from mathutils import Vector

pattern_presets = {  # list of points position that make 2d bezier curve pattern
    'SHIRT': {
        'CO': [(0.0, 0.87, 0.0), (0.24, 1.04, 0.0), (0.61, 0.92, 0.0), (0.71, 0.29, 0.0), (0.72, -0.54, 0.0), (0.0, -0.54, 0.0), (-0.72, -0.54, 0.0), (-0.71, 0.29, 0.0), (-0.61, 0.92, 0.0), (-0.24, 1.04, 0.0)] ,
        'HL': [(-0.11, 0.87, 0.0), (0.21, 0.94, 0.0), (0.49, 0.96, 0.0), (0.48, 0.43, 0.0), (0.72, -0.26, 0.0), (0.24, -0.54, 0.0), (-0.48, -0.54, 0.0), (-0.71, 0.01, 0.0), (-0.57, 0.73, 0.0), (-0.36, 1.0, 0.0)] ,
        'HL_TYPE': ['AUTO', 'FREE', 'VECTOR', 'FREE', 'VECTOR', 'FREE', 'FREE', 'VECTOR', 'FREE', 'VECTOR'] ,
        'HR': [(0.11, 0.87, 0.0), (0.36, 1.0, 0.0), (0.57, 0.73, 0.0), (0.71, 0.01, 0.0), (0.48, -0.54, 0.0), (-0.24, -0.54, 0.0), (-0.72, -0.26, 0.0), (-0.48, 0.43, 0.0), (-0.49, 0.96, 0.0), (-0.21, 0.94, 0.0)] ,
        'HR_TYPE': ['AUTO', 'VECTOR', 'FREE', 'VECTOR', 'FREE', 'FREE', 'VECTOR', 'FREE', 'VECTOR', 'FREE'] ,
    },
    'SLEEVE': {
        'CO': [(0.0, 0.4, 0.0), (0.56, -0.01, 0.0), (0.45, -1.32, 0.0), (0.0, -1.32, 0.0), (-0.45, -1.32, 0.0), (-0.56, -0.01, 0.0)] ,
        'HL': [(-0.27, 0.4, 0.0), (0.32, 0.11, 0.0), (0.49, -0.89, 0.0), (0.18, -1.32, 0.0), (-0.3, -1.32, 0.0), (-0.52, -0.45, 0.0)] ,
        'HL_TYPE': ['AUTO', 'FREE', 'VECTOR', 'AUTO', 'VECTOR', 'VECTOR'] ,
        'HR': [(0.27, 0.4, 0.0), (0.52, -0.45, 0.0), (0.3, -1.32, 0.0), (-0.18, -1.32, 0.0), (-0.49, -0.89, 0.0), (-0.32, 0.11, 0.0)] ,
        'HR_TYPE': ['AUTO', 'VECTOR', 'VECTOR', 'AUTO', 'VECTOR', 'FREE'] ,
    },
    'PANTS': {
        'CO': [(0.26, 1.46, 0.0), (1.06, 1.47, 0.0), (0.97, -1.0, 0.0), (0.2, -1.0, 0.0), (0.04, 0.59, 0.0)] ,
        'HL': [(0.26, 1.17, 0.0), (0.79, 1.47, 0.0), (1.04, -0.09, 0.0), (0.2, -1.0, 0.0), (0.22, 0.34, 0.0)] ,
        'HL_TYPE': ['FREE', 'VECTOR', 'FREE', 'FREE', 'FREE'] ,
        'HR': [(0.53, 1.46, 0.0), (1.06, 0.64, 0.0), (0.71, -1.0, 0.0), (0.2, -0.37, 0.0), (0.25, 0.67, 0.0)] ,
        'HR_TYPE': ['VECTOR', 'FREE', 'VECTOR', 'FREE', 'FREE'] ,
    },
    'PLANE': {
        'CO': [Vector((0.5, -0.5, 0.0)), Vector((0.5, 0.5, 0.0)), Vector((-0.5, 0.5, 0.0)), Vector((-0.5, -0.5, 0.0))],
        'HL': [Vector((0.5, -0.5, 0.0)), Vector((0.5, 0.5, 0.0)), Vector((-0.5, 0.5, 0.0)), Vector((-0.5, -0.5, 0.0))],
        'HL_TYPE': ['VECTOR', 'VECTOR', 'VECTOR', 'VECTOR'],
        'HR': [Vector((0.5, -0.5, 0.0)), Vector((0.5, 0.5, 0.0)), Vector((-0.5, 0.5, 0.0)), Vector((-0.5, -0.5, 0.0))],
        'HR_TYPE': ['VECTOR', 'VECTOR', 'VECTOR', 'VECTOR']
    }
}


class GTOOL_OT_Add2DBezierCurve(bpy.types.Operator):
    """Add a 2D Bezier Curve with field and both sides"""
    bl_idname = "curve.add_sewing_pattern"
    bl_label = "Add sewing pattern"
    bl_description = "Add sewing pattern preset -  2D Bezier Curve"
    bl_options = {'REGISTER', 'UNDO'}

    pattern_type: bpy.props.EnumProperty(
        name="Pattern Type",
        items=[
            ('PLANE', "Plane", "Plane"),
            ('SHIRT', "Shirt", "Shirt"),
            ('SLEEVE', "Sleeve", "Sleeve"),
            ('PANTS', "Pants", "Pants"),
        ],
        default='SHIRT',
    )

    def execute(self, context):
        # Create a new Bezier curve
        curve = bpy.data.curves.new(name=self.pattern_type, type='CURVE')
        curve.splines.new('BEZIER')
        curve.dimensions = '2D'
        curve.fill_mode = 'BOTH'

        # curve.bevel_depth = 0.1
        # curve.bevel_resolution = 4
        # curve.render_resolution = 20

        # Create a new object with the curve
        curve_obj = bpy.data.objects.new(self.pattern_type, curve)

        context.collection.objects.link(curve_obj)
        pattern_data = pattern_presets[self.pattern_type]
        spline = curve.splines[0]
        spline.use_cyclic_u = True
        spline.bezier_points.add(count=len(pattern_data['CO']) - 1)
        for i, co in enumerate(pattern_data['CO']):
            spline.bezier_points[i].co = co
            spline.bezier_points[i].handle_left_type = pattern_data['HL_TYPE'][i]
            spline.bezier_points[i].handle_right_type = pattern_data['HR_TYPE'][i]
            spline.bezier_points[i].handle_left = pattern_data['HL'][i]
            spline.bezier_points[i].handle_right = pattern_data['HR'][i]

        for obj in context.selected_objects:
            obj.select_set(False)
        curve_obj.select_set(True)
        context.view_layer.objects.active = curve_obj

        return {'FINISHED'}
