
'''
Copyright (C) 2017 JOSECONSCO
Created by JOSECONSCO

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import bpy
import gpu
import copy
from gpu_extras.batch import batch_for_shader
from mathutils import Vector, Matrix
from mathutils.bvhtree import BVHTree
from mathutils.geometry import intersect_line_sphere, intersect_point_line
from bpy_extras import view3d_utils
import numpy as np
from ..utils.helper_functions import draw_text_line, view_obj_raycast, get_weights, set_weights
from ..utils.fsm_oper import ModalCommon
from .interaction_helpers import *

#DONE: check if cloth is using pin group already, if so use it for dragging. Restore previous vertex pin weight
#DONE: play only when clicked
#DONE: reset keyframes on loop?
#DONE: badd pulling direction if modifiers are enabled??? Somewehre they are disabled, then pull works ok. Disable moddsss..
#TODO: removing rest_shape, base_shape breaks the sim? buggy cloth cache? #! RESET  BLENDER HELPS

FRAME_CHANGED = False  # True on  handlers.frame_change_post
LAST_FRAME_SHAPE = None  # Cache cloth on scene.frame_end
FIRST_RUN = True #true if this is first animation loop playback
DEPSGRAPH = False

CLOTH_OBJ = None

def mute_animations(mute = True):
    if bpy.context.scene.animation_data and bpy.context.scene.animation_data.action:
        for fcurve in bpy.context.scene.animation_data.action.fcurves:
            fcurve.mute = mute
    if bpy.context.active_object.animation_data and bpy.context.active_object.animation_data.action:
        for fcurve in bpy.context.active_object.animation_data.action.fcurves:
            fcurve.mute = mute

def enable_back_mods():
    '''Basically to enable bend mod of last frame..'''
    mesh_objs = [obj for obj in bpy.context.scene.objects if obj.type == 'MESH']
    objs_with_cloth = [obj for obj in mesh_objs if any([mod.type == 'CLOTH' for mod in obj.modifiers if mod.show_viewport])]
    for cloth_obj in objs_with_cloth:
        for mod in cloth_obj.modifiers:
            if mod.type != 'CLOTH':
                mod.show_viewport = True



def frame_change_update(dummy):
    global LAST_FRAME_SHAPE, FIRST_RUN, DEPSGRAPH248
    cloth_obj = CLOTH_OBJ
    # scene = bpy.context.scene
    # if scene.garment_props.continuous_sim:
    #     # NOTE: restore last frame mesh shape at loop
    #     if scene.frame_current == scene.frame_end:  # backup cloth to LAST_FRAME_SHAPE
    #         # print('last frame foreach_get')
    #         vert_count = len(cloth_obj.data.vertices)
    #         LAST_FRAME_SHAPE = np.empty((vert_count*3), 'f')
    #         DEPSGRAPH.update()
    #         cloth_eval = cloth_obj.evaluated_get(DEPSGRAPH) #cloth obj, with modifiers
    #         cloth_eval.data.vertices.foreach_get("co", LAST_FRAME_SHAPE)
    #         if FIRST_RUN:
    #             FIRST_RUN = False
    #             mute_animations(True)
    #     if scene.frame_current == scene.frame_start:  # reset sim_shape [0] to last sim shape - LAST_FRAME_SHAPE
    #         if LAST_FRAME_SHAPE is not None and 'rest_shape' in cloth_obj.data.shape_keys.key_blocks:
    #             # print('first frame foreach_set')
    #             vert_count = len(cloth_obj.data.vertices)
    #             rest_shape_co = np.empty((vert_count*3), 'f')
    #             cloth_obj.data.shape_keys.key_blocks['rest_shape'].data.foreach_get("co", rest_shape_co)
    #             #works without pin, cos first frame? Won't work with finflate...
    #             cloth_obj.data.shape_keys.key_blocks[0].data.foreach_set("co", LAST_FRAME_SHAPE)
    #             cloth_obj.data.shape_keys.key_blocks[0].data.update()
    #             cloth_obj.data.vertices.foreach_set("co", LAST_FRAME_SHAPE)# no need for seem, but fixes ugly jump of pulled verts, on sim
    #             cloth_obj.data.update()
    # else:
    #     # NOTE: restore last frame mesh shape at loop
    #     if scene.frame_current == scene.frame_start:  # reset sim_shape [0] to 'rest_shape'
    #         if 'rest_shape' in cloth_obj.data.shape_keys.key_blocks:
    #             # print('frame_change_update: resseting rest_shape \n \n')
    #             vert_count = len(cloth_obj.data.vertices)
    #             rest_shape_co = np.empty((vert_count*3), dtype=np.float)
    #             cloth_obj.data.shape_keys.key_blocks['rest_shape'].data.foreach_get("co", rest_shape_co)
    #             cloth_obj.data.shape_keys.key_blocks[0].data.foreach_set("co", rest_shape_co)
    #             cloth_obj.data.shape_keys.key_blocks[0].data.update()
    #             cloth_obj.data.vertices.foreach_set("co", rest_shape_co)  # no need for seem, but fixes ugly jump of pulled verts, on sim
    #             cloth_obj.data.update()

    # if scene.frame_current == scene.frame_end:  # reset sim_shape [0] to 'rest_shape'
    #     # enable mods on loop, to prevent 2 times influence (eg. baked shape bend + mod bend), also mods break pulling too.
    #     enable_back_mods()

    global FRAME_CHANGED
    FRAME_CHANGED = True


class GTOOL_OT_CreateRestShape(bpy.types.Operator):
    bl_idname = "cloth.create_rest_shape"
    bl_label = "Create/Update rest shape"
    bl_description = "Required for cloth interaction to work.\nCreated automatically when no rest_shape is detected \nUsually it should be used after you modify cloth mesh topology/shape"
    bl_options = {"REGISTER", "UNDO"}

    @classmethod
    def poll(cls, context):
        return context.active_object and context.active_object.type == 'MESH'

    def execute(self, context):
        cloth_obj = context.active_object
        create_rest_shape(context, cloth_obj)
        pin_vg, pin_existed = get_pin_vertex_group(cloth_obj)
        return {"FINISHED"}


class GTOOL_OT_ResetClothSim(bpy.types.Operator):
    bl_idname = "cloth.reset_cloth_sim"
    bl_label = "Reset cloth simulation"
    bl_description = "Reset cloth simulation. Restore mesh from 'rest_shape'"
    bl_options = {"REGISTER", "UNDO"}

    @classmethod
    def poll(cls, context):
        return context.active_object and context.active_object.type == 'MESH'

    def execute(self, context):
        cloth_obj = context.active_object
        if cloth_obj.data.shape_keys and 'rest_shape' in cloth_obj.data.shape_keys.key_blocks.keys():

            rest_shape = cloth_obj.data.shape_keys.key_blocks['rest_shape']
            base_shape = cloth_obj.data.shape_keys.key_blocks[0]
            base_shape.name = 'sim_shape'

            rest_shape_co = np.zeros((len(cloth_obj.data.vertices)*3), dtype=np.float)
            rest_shape.data.foreach_get("co", rest_shape_co)

            base_shape.data.foreach_set("co", rest_shape_co)
            base_shape.data.update()

            cloth_obj.data.vertices.foreach_set("co", rest_shape_co)  # noo nedd for seem, but fixes ugly jump of pulled verts, on sim
            cloth_obj.data.update()

            bpy.context.scene.frame_current = bpy.context.scene.frame_start
            global LAST_FRAME_SHAPE
            LAST_FRAME_SHAPE = None
        else:
            print('Cant reset selected object. No "rest_shape" found!')
        for mod in cloth_obj.modifiers:
            if mod.type != 'CLOTH':
                mod.show_viewport = True
        mute_animations(False)

        return {"FINISHED"}


class GTOOL_OT_BakeSimToShape(bpy.types.Operator):
    bl_idname = "cloth.save_to_shape"
    bl_label = "Save to shape key"
    bl_description = "Save simulation result to new shape key"
    bl_options = {"REGISTER", "UNDO"}

    @classmethod
    def poll(cls, context):
        return context.mode == "OBJECT"

    def execute(self, context):
        cloth_obj = context.active_object
        if not cloth_obj.data.shape_keys:
            bpy.ops.object.shape_key_add(from_mix=False)  # add  base key
        # for i, sourceShape in enumerate(cloth_obj.data.shape_keys.key_blocks):
        #     sourceShape.value = 0
        cloth_obj.shape_key_add(name='Stored Sim', from_mix=False)
        target_shape = cloth_obj.data.shape_keys.key_blocks[-1]

        constructive_modifiers = ['SUBSURF', 'BEVEL', 'MIRROR', 'MASK', 'EDGE_SPLIT', 'SOLIDIFY']
        disabled_mods = []
        for mod in cloth_obj.modifiers:  # disable constructive modifiers - cant write to shape if new geo is created
            if mod.type in constructive_modifiers and not mod.show_viewport:
                disabled_mods.append(mod)
                mod.show_viewport = False
        depsgraph = bpy.context.evaluated_depsgraph_get()
        cloth_obj_eval = cloth_obj.evaluated_get(depsgraph)
        for meshVert in cloth_obj_eval.data.vertices:
            target_shape.data[meshVert.index].co = meshVert.co

        cloth_obj.data.shape_keys.key_blocks[-1].value = 1
        for mod in cloth_obj.modifiers:  # disable all mods
            mod.show_viewport = False
        # for mod in disabled_mods:  # disable constructive modifiers
        #     mod.show_viewport = True
        return {"FINISHED"}



class GTOOL_OT_PullCloth(bpy.types.Operator, ModalCommon): # , CookieCutter_FSM):
    """Modal object selection with a ray cast"""
    bl_idname = "cloth.pull_cloth"
    bl_label = "Interact with cloth"
    bl_description = "Pull cloth object with LMB while it is simulated in 3d viewport"
    bl_options = {'REGISTER', 'UNDO'}

    @staticmethod
    def draw_text(self, context):
        draw_text_lines = []
        state_txt = "LMB - pull cloth"if self.state == 'main' else 'Pulling'
        draw_text_lines.append([state_txt, 'H1'])
        draw_text_lines.append([f"(R)eset sim", 'p'])
        overlays = 'ON' if self.draw_overlays else 'OFF'
        draw_text_lines.append([f"(D)raw overlays: {overlays}", 'p'])
        draw_text_lines.append(["End: RMB\SPACE", 'p'])
        draw_text_line(draw_text_lines)

    @staticmethod
    def zero_sim():
        bpy.context.scene.frame_current = bpy.context.scene.frame_start
        global LAST_FRAME_SHAPE
        LAST_FRAME_SHAPE = None


    def end_cancel(self):
        cloth_obj = self.cloth_obj
        enable_back_mods()
        if self.pin_vg_weights_backup:#reest weights
            set_weights(cloth_obj, self.pin_vg_weights_backup, self.pin_vg)
        self.zero_sim()
        if self.empty:
            bpy.data.objects.remove(self.empty)

        if frame_change_update in bpy.app.handlers.frame_change_post:
            bpy.app.handlers.frame_change_post.remove(frame_change_update)
        self.disable_handlers()
        if bpy.context.screen.is_animation_playing:
            bpy.ops.screen.animation_play()
        return {'CANCELLED'}

    def end_finish(self):
        enable_back_mods()
        if frame_change_update in bpy.app.handlers.frame_change_post:
            bpy.app.handlers.frame_change_post.remove(frame_change_update)
        self.disable_handlers()
        if self.empty:
            bpy.data.objects.remove(self.empty)
        if bpy.context.screen.is_animation_playing:
            bpy.ops.screen.animation_play()
        return {'FINISHED'}

    coordinates = []
    @staticmethod
    def draw_px(self, context):
        if self.draw_overlays:
            gpu.state.blend_set("ALPHA")
            gpu.state.point_size_set(8)
            gpu.state.depth_test_set("LESS")

            shader = self.shader
            if self.state == 'drag_empty' and self.drag_target_loc:
                batch = batch_for_shader(shader, "POINTS", {"pos": [self.drag_target_loc]})
                shader.bind()
                shader.uniform_float("color", (1.0, 0.7, 0.2, 0.9))
                batch.draw(shader)
            if self.pinned_verts_ids and FRAME_CHANGED:
                cloth_eval = context.active_object.evaluated_get(DEPSGRAPH)
                self.coordinates = [context.active_object.matrix_world@cloth_eval.data.vertices[v_id].co for v_id in self.pinned_verts_ids]
            if self.pinned_verts_ids:
                batch = batch_for_shader(shader, "POINTS", {"pos": self.coordinates})
                shader.bind()
                shader.uniform_float("color", (1.0, 0.3, 0.0, 0.7))
                batch.draw(shader)

            #restore opengl defaults
            gpu.state.depth_test_set("NONE")
            gpu.state.point_size_set(1)
            gpu.state.blend_set("NONE")
        return

    def get_closest_vert_id(self, obj, hit_face_id, hit_loc_local, depsgraph):
        ''' Get face vert, that is closetst to hit_loc_local, from eval_mesh '''
        hit_face = obj.data.polygons[hit_face_id]
        smallest_dist = 100
        best_vert_id = None
        best_vert_co = None
        obj_eval = obj.evaluated_get(depsgraph)
        for v_id in hit_face.vertices:
            new_dist = (obj_eval.data.vertices[v_id].co-hit_loc_local).length_squared
            if new_dist < smallest_dist:
                best_vert_id = v_id
                smallest_dist = new_dist
                best_vert_co = copy.deepcopy(obj_eval.data.vertices[v_id].co)
        return best_vert_id, best_vert_co

    @staticmethod
    def add_empty_with_field(context, location):
        scn_gt_props = context.scene.garment_props
        empty = bpy.data.objects.new('GTOOL_CLOTH_PULL', None)
        empty.empty_display_type = 'SPHERE'
        empty.empty_display_size = 0.01 * context.active_object.dimensions.length
        empty.location = location
        # empty.field = bpy.data.fields.new(name='GTOOL_CLOTH_PULL') # not working (no filed dataytpe)
        # use bpy.ops.object.forcefield_toggle() , with context override
        override = {'active_object': empty, 'selected_objects': [empty], 'object': empty, 'selected_editable_objects': [empty], 'visible_objects': [empty], 'editable_objects': [empty]}
        with bpy.context.temp_override(**override):
            bpy.ops.object.forcefield_toggle()

        empty.field.type = 'FORCE'
        empty.field.strength = -10000*scn_gt_props.pull_force # in up to 10
        empty.field.apply_to_location = True
        empty.field.apply_to_rotation = False

        empty.field.falloff_power = 10*(1-scn_gt_props.pull_falloff)
        empty.field.use_max_distance = scn_gt_props.use_max_distance
        empty.field.distance_max = scn_gt_props.distance_max
        empty.field.use_min_distance = False
        # empty.field.distance_min = 0.01
        return empty

    def drag_move_set_vert_co(self, context, event):
        """calculate moved vert co from cursor to location 3d, execute the ray cast"""
        coord = event.mouse_region_x, event.mouse_region_y
        depth_location = self.cloth_obj.matrix_world @ self.best_hit_result
        self.drag_target_loc = view3d_utils.region_2d_to_location_3d(context.region, context.region_data, coord, depth_location)
        drag_target_loc_world = self.mw_inv @ self.drag_target_loc
        self.empty.location = drag_target_loc_world


    def modal_main(self, context, event):
        cloth_obj = self.cloth_obj
        if event.type == 'LEFTMOUSE' and event.value == 'PRESS':
            # print('main: Picked surface point')
            # disable mods on loop, to prevent 2 times influence (eg. baked shape bend + mod bend), also mods break pulling too.
            for mod in cloth_obj.modifiers:
                if mod.type != 'CLOTH':
                    mod.show_viewport = False
            hit_loc_local, normal, face_index, dist = view_obj_raycast(context,  cloth_obj, event, applyModifiers=True, depsgraph=DEPSGRAPH, world_space=False)
            if hit_loc_local:
                self.hit_vert_id, self.best_hit_result = self.get_closest_vert_id(cloth_obj, face_index, hit_loc_local, DEPSGRAPH)
                self.empty = self.add_empty_with_field(context, location=self.best_hit_result)
                context.collection.objects.link(self.empty)

                try:  # when we release dragged vert restore weight to - previous_vert_weight
                    self.previous_vert_weight = self.pin_vg.weight(self.hit_vert_id)
                except:
                    self.previous_vert_weight = 0
                # if not context.screen.is_animation_playing and self.payback_method == 'INTERRUPTED':
                #     bpy.ops.screen.animation_play()
                self.state = 'drag_empty'
            else:
                self.report({'INFO'}, 'Click on mesh pattern')

        elif event.type == 'D' and event.value == 'PRESS':
            self.draw_overlays = not self.draw_overlays

        elif event.type == 'R' and event.value == 'PRESS': #reset sim
            self.zero_sim()


    def drag_empty(self, context, event):
        global FRAME_CHANGED
        if FRAME_CHANGED and event.type == 'MOUSEMOVE':
            self.drag_move_set_vert_co(context, event)
            FRAME_CHANGED = False

        elif event.type == 'LEFTMOUSE' and event.value == 'RELEASE':  # should only be executed on LMB
            bpy.data.objects.remove(self.empty) # New way
            self.empty = None
            # if context.screen.is_animation_playing and self.payback_method == 'INTERRUPTED':
            #     bpy.ops.screen.animation_play()
            self.state = 'main'



    def modal(self, context, event):
        context.area.tag_redraw()

        #reset sim if non looped playback
        # global FRAME_CHANGED
        # if not context.scene.garment_props.continuous_sim and context.scene.frame_current == context.scene.frame_end and FRAME_CHANGED:
        #     FRAME_CHANGED = False

        if self.state == 'main':
            self.modal_main(context, event)
        elif self.state == 'drag_empty':
            self.drag_empty(context, event)

        if event.type in {'MIDDLEMOUSE', 'WHEELUPMOUSE', 'WHEELDOWNMOUSE'}:
            return {'PASS_THROUGH'}

        elif event.type in {'RIGHTMOUSE', 'RET', 'SPACE'} and event.value == 'PRESS':
            return self.end_finish()

        elif event.type in {'ESC'} and event.value == 'PRESS':
            return self.end_finish()
            # return self.end_cancel()

        return {'RUNNING_MODAL'}
        if event.alt and event.type == 'q' and event.value == 'PRESS': #FORCE QUIT
            return {"CANCELLED"}


    def before_start(self, context, event):
        global FIRST_RUN, DEPSGRAPH, CLOTH_OBJ
        FIRST_RUN = True
        self.draw_overlays = False
        self.state = 'main' # initial state : 'main', 'drag_empty',
        cloth_obj = context.active_object
        CLOTH_OBJ = cloth_obj
        self.cloth_obj = cloth_obj

        self.pin_vg, self.pin_vg_existed = get_pin_vertex_group(cloth_obj)
        if self.pin_vg_existed:  # do we created pin_vg, or we used existing one from cloth_mod.vertex_group_mass
            self.pin_vg_weights_backup = get_weights(cloth_obj, self.pin_vg) # for resetting pin_vg to inital weights
        else:
            self.pin_vg_weights_backup = []

        self.empty = None # force field empty
        self.mw_inv = cloth_obj.matrix_world.inverted() #just for performance
        self.hit_vert_id = None # id of picked vert with LMB
        self.drag_target_loc = None #helper for gpu drawing only
        self.all_vert_ids = [i for i in range(len(cloth_obj.data.vertices))] #another cache for all verts list
        # self.payback_method = context.scene.garment_props.animation_method  # CONSTANT of INTERRUPTED
        if not bpy.context.screen.is_animation_playing: # and self.payback_method == 'CONSTANT'
            bpy.ops.screen.animation_play()

        if frame_change_update not in bpy.app.handlers.frame_change_post:
            bpy.app.handlers.frame_change_post.append(frame_change_update)

        DEPSGRAPH = context.evaluated_depsgraph_get()
        return

    def invoke(self, context, event):
        if context.space_data.type == 'VIEW_3D':
            if not context.active_object or context.active_object.type != 'MESH' or not any([mod.type == 'CLOTH' for mod in context.active_object.modifiers]):
                self.report({'WARNING'}, "Select mesh object first, with cloth modifier. Cancelling")
                return {'CANCELLED'}
            if any([mod.type == 'CLOTH' and mod.point_cache.is_baked for mod in context.active_object.modifiers]):
                self.report({'INFO'}, f'Object {context.active_object.name} is cached! Disable clot cache for interaction. Cancelling')
                return {'CANCELLED'}
            self.invoke_basic(context, event)
            # self.fsm_init()
            self.before_start(context, event)
            context.window_manager.modal_handler_add(self)
            return {'RUNNING_MODAL'}
        else:
            self.report({'WARNING'}, "Active space must be a View3d")
            return {'CANCELLED'}



# old ver
# class GTOOL_OT_PullCloth(bpy.types.Operator, ModalCommon): # , CookieCutter_FSM):
#     """Modal object selection with a ray cast"""
#     bl_idname = "cloth.pull_cloth"
#     bl_label = "Interact with cloth"
#     bl_description = "Pull cloth object with LMB while it is simulated in 3d viewport"
#     bl_options = {'REGISTER', 'UNDO'}
#
#     @staticmethod
#     def draw_text(self, context):
#         draw_text_lines = []
#         state_txt = "LMB - pull cloth"if self.state == 'main' else 'Pulling'
#         draw_text_lines.append([state_txt, 'H1'])
#         pining = 'ON' if self.with_pinning else 'OFF'
#         draw_text_lines.append([f"(P)in: {pining}", 'p'])
#         draw_text_lines.append(["(U)npin", 'p'])
#         draw_text_lines.append([f"(C)lear (All)pins:  (SHIFT)+C", 'p'])
#         draw_text_lines.append([f"(R)eset sim", 'p'])
#         overlays = 'ON' if self.draw_overlays else 'OFF'
#         draw_text_lines.append([f"(D)raw overlays: {overlays}", 'p'])
#         draw_text_lines.append(["End: RMB\SPACE", 'p'])
#         draw_text_line(draw_text_lines)
#
#     def end_cancel(self):
#         cloth_obj = self.cloth_obj
#         if self.pin_vg_weights_backup:#reest weights
#             set_weights(cloth_obj, self.pin_vg_weights_backup, self.pin_vg)
#         bpy.ops.cloth.reset_cloth_sim() #reset sim
#
#         if frame_change_update in bpy.app.handlers.frame_change_post:
#             bpy.app.handlers.frame_change_post.remove(frame_change_update)
#         self.disable_handlers()
#         if bpy.context.screen.is_animation_playing:
#             bpy.ops.screen.animation_play()
#         return {'CANCELLED'}
#
#     def end_finish(self):
#         if frame_change_update in bpy.app.handlers.frame_change_post:
#             bpy.app.handlers.frame_change_post.remove(frame_change_update)
#         self.disable_handlers()
#         if bpy.context.screen.is_animation_playing:
#             bpy.ops.screen.animation_play()
#         return {'FINISHED'}
#
#     coordinates = []
#     @staticmethod
#     def draw_px(self, context):
#         if self.draw_overlays:
#             gpu.state.blend_set("ALPHA")
#             gpu.state.point_size_set(8)
#             gpu.state.depth_test_set("LESS")
#
#             shader = self.shader
#             if self.state == 'drag_empty' and self.drag_target_loc:
#                 batch = batch_for_shader(shader, "POINTS", {"pos": [self.drag_target_loc]})
#                 shader.bind()
#                 shader.uniform_float("color", (1.0, 0.7, 0.2, 0.9))
#                 batch.draw(shader)
#             if self.pinned_verts_ids and FRAME_CHANGED:
#                 cloth_eval = context.active_object.evaluated_get(DEPSGRAPH)
#                 self.coordinates = [context.active_object.matrix_world@cloth_eval.data.vertices[v_id].co for v_id in self.pinned_verts_ids]
#             if self.pinned_verts_ids:
#                 batch = batch_for_shader(shader, "POINTS", {"pos": self.coordinates})
#                 shader.bind()
#                 shader.uniform_float("color", (1.0, 0.3, 0.0, 0.7))
#                 batch.draw(shader)
#
#             #restore opengl defaults
#             gpu.state.depth_test_set("NONE")
#             gpu.state.point_size_set(1)
#             gpu.state.blend_set("NONE")
#         return
#
#     def get_closest_pinned_vert_id(self, context,  event, cloth_obj, depsgraph):
#         ''' Get face vert, that is closetst to hit_loc_local, from eval_mesh '''
#         region = context.region
#         rv3d = context.region_data
#         coord = event.mouse_region_x, event.mouse_region_y
#         # coord = mouse_coordinates
#
#         # get the ray from the viewport and mouse
#         view_vector = view3d_utils.region_2d_to_vector_3d(region, rv3d, coord)
#         ray_origin = view3d_utils.region_2d_to_origin_3d(region, rv3d, coord)
#
#         matrix_inv = self.mw_inv  # faster than transforming whole mesh
#         ray_origin_obj = matrix_inv @ ray_origin
#         view_vector_obj = matrix_inv @ Vector(view_vector.xyz[:]+(0,)) #make 4d vec.
#         view_vector_obj = view_vector_obj.to_3d() # back to 3d
#
#         smallest_dist = 1000
#         best_vert_id = None
#         cloth_obj_eval = cloth_obj.evaluated_get(depsgraph)
#         for v_id in self.pinned_verts_ids:
#             vert = cloth_obj_eval.data.vertices[v_id]
#             closest_to_line, factor_ray_origin = intersect_point_line(vert.co, ray_origin_obj, ray_origin_obj + view_vector_obj)
#             new_dist = (vert.co-closest_to_line).length_squared
#             if new_dist < smallest_dist:
#                 best_vert_id = v_id
#                 smallest_dist = new_dist
#         return best_vert_id
#
#     def get_closest_vert_id(self, obj, hit_face_id, hit_loc_local, depsgraph):
#         ''' Get face vert, that is closetst to hit_loc_local, from eval_mesh '''
#         hit_face = obj.data.polygons[hit_face_id]
#         smallest_dist = 100
#         best_vert_id = None
#         best_vert_co = None
#         obj_eval = obj.evaluated_get(depsgraph)
#         for v_id in hit_face.vertices:
#             new_dist = (obj_eval.data.vertices[v_id].co-hit_loc_local).length_squared
#             if new_dist < smallest_dist:
#                 best_vert_id = v_id
#                 smallest_dist = new_dist
#                 best_vert_co = copy.deepcopy(obj_eval.data.vertices[v_id].co)
#         return best_vert_id, best_vert_co
#
#     @staticmethod
#     def add_empty_with_field(context, location):
#         scn_gt_props = context.scene.garment_props
#         empty = bpy.data.objects.new('GTOOL_CLOTH_PULL', None)
#         empty.empty_display_type = 'SPHERE'
#         empty.empty_display_size = 0.1
#         empty.location = location
#         # empty.field = bpy.data.fields.new(name='GTOOL_CLOTH_PULL') # not working (no filed dataytpe)
#         # use bpy.ops.object.forcefield_toggle() , with context override
#         override = {'active_object': empty, 'selected_objects': [empty], 'object': empty, 'selected_editable_objects': [empty], 'visible_objects': [empty], 'editable_objects': [empty]}
#         with bpy.context.temp_override(**override):
#             bpy.ops.object.forcefield_toggle()
#
#         empty.field.type = 'FORCE'
#         empty.field.strength = -2000*scn_gt_props.pull_force # in up to 10
#         empty.field.apply_to_location = True
#         empty.field.apply_to_rotation = False
#
#         empty.field.falloff_power = 10*scn_gt_props.pull_falloff
#         empty.field.use_max_distance = False
#         empty.field.distance_max = 0.5
#         empty.field.use_min_distance = False
#         # empty.field.distance_min = 0.01
#         return empty
#
#     def drag_move_set_vert_co(self, context, event):
#         """calculate moved vert co from cursor to location 3d, execute the ray cast"""
#         coord = event.mouse_region_x, event.mouse_region_y
#         cloth_obj = self.cloth_obj
#         depth_location = cloth_obj.matrix_world @ self.best_hit_result
#         self.drag_target_loc = view3d_utils.region_2d_to_location_3d(context.region, context.region_data, coord, depth_location)
#         drag_target_loc_world = self.mw_inv @ self.drag_target_loc
#
#         # self.best_hit_result = self.best_hit_result.lerp(drag_target_loc_world, self.pull_strenght)  # ease to
#         # cloth_obj.data.shape_keys.key_blocks[0].data[self.hit_vert_id].co = drag_target_loc_world # XXX: old way - worked in B 3.3
#         # TODO: investigate - reset shape key is  not maintained (check maybe force reset it). Then empty forcefield - coiuld be used as second option
#         self.empty.location = drag_target_loc_world
#
#
#     def modal_main(self, context, event):
#         cloth_obj = self.cloth_obj
#         if event.type == 'LEFTMOUSE' and event.value == 'PRESS':
#             # print('main: Picked surface point')
#             # disable mods on loop, to prevent 2 times influence (eg. baked shape bend + mod bend), also mods break pulling too.
#             for mod in cloth_obj.modifiers:
#                 if mod.type != 'CLOTH':
#                     mod.show_viewport = False
#             hit_loc_local, normal, face_index, dist = view_obj_raycast(context,  cloth_obj, event, applyModifiers=True, depsgraph=DEPSGRAPH, world_space=False)
#             if hit_loc_local:
#                 self.hit_vert_id, self.best_hit_result = self.get_closest_vert_id(cloth_obj, face_index, hit_loc_local, DEPSGRAPH)
#                 self.empty = self.add_empty_with_field(context, location=self.best_hit_result)
#                 context.collection.objects.link(self.empty)
#
#                 try:  # when we release dragged vert restore weight to - previous_vert_weight
#                     self.previous_vert_weight = self.pin_vg.weight(self.hit_vert_id)
#                 except:
#                     self.previous_vert_weight = 0
#                 self.pull_strenght = 0.1
#                 if not context.screen.is_animation_playing and self.payback_method == 'INTERRUPTED':
#                     bpy.ops.screen.animation_play()
#                 self.state = 'drag_empty'
#             else:
#                 self.report({'INFO'}, 'Click on mesh pattern')
#
#         elif event.type == 'P' and event.value == 'PRESS':
#             self.with_pinning = not self.with_pinning
#
#         elif event.type == 'D' and event.value == 'PRESS':
#             self.draw_overlays = not self.draw_overlays
#
#
#         elif event.type == 'U' and event.value == 'PRESS':
#             hit_vert_id = self.get_closest_pinned_vert_id(context,  event, cloth_obj, DEPSGRAPH)
#             if hit_vert_id:
#                 self.pinned_verts_ids.remove(hit_vert_id)
#                 self.pin_vg.remove([hit_vert_id])
#
#
#         elif event.type == 'R' and event.value == 'PRESS': #reset sim
#             set_weights(cloth_obj, self.pin_vg_weights_backup, self.pin_vg)
#             self.pinned_verts_ids = [v_id for v_id, weight in enumerate(self.pin_vg_weights_backup) if weight and weight[0]]
#             bpy.ops.cloth.reset_cloth_sim()
#
#         elif event.type == 'C' and event.value == 'PRESS': #"C - Clear pinning"
#             if event.shift: #clear all pinned weights
#                 self.pin_vg.remove(self.all_vert_ids)
#                 self.pinned_verts_ids.clear()
#             else:
#                 # reset to cached  pin_vg_weights_backup
#                 set_weights(cloth_obj, self.pin_vg_weights_backup, self.pin_vg)
#                 self.pinned_verts_ids = [v_id for v_id, weight in enumerate(self.pin_vg_weights_backup) if weight and weight[0]]
#
#
#         elif event.type in {'MIDDLEMOUSE', 'WHEELUPMOUSE', 'WHEELDOWNMOUSE'}:
#             return {'PASS_THROUGH'}
#
#         elif event.type in {'RIGHTMOUSE', 'RET', 'SPACE'} and event.value == 'PRESS':
#             return self.end_finish()
#
#         elif event.type in {'ESC'} and event.value == 'PRESS':
#             return self.end_cancel()
#
#         #reset sim if non looped playback
#         global FRAME_CHANGED
#         if not context.scene.garment_props.continuous_sim and context.scene.frame_current == context.scene.frame_end and FRAME_CHANGED:
#             set_weights(cloth_obj, self.pin_vg_weights_backup, self.pin_vg)
#             self.pinned_verts_ids = [v_id for v_id, weight in enumerate(self.pin_vg_weights_backup) if weight and weight[0]]
#             FRAME_CHANGED = False
#
#         return {'RUNNING_MODAL'}
#
#
#     def drag_empty(self, context, event):
#         global FRAME_CHANGED
#
#         #reset sim if non looped playback
#         if not context.scene.garment_props.continuous_sim and context.scene.frame_current == context.scene.frame_end and FRAME_CHANGED:
#             set_weights(self.cloth_obj, self.pin_vg_weights_backup, self.pin_vg)
#             self.pinned_verts_ids = [v_id for v_id, weight in enumerate(self.pin_vg_weights_backup) if weight and weight[0]]
#             FRAME_CHANGED = False
#
#         if FRAME_CHANGED and event.type == 'MOUSEMOVE':
#             # self.pin_vg.add([self.hit_vert_id], self.pull_strenght, "REPLACE") # XXX: old way - worked in B 3.3
#             self.pull_strenght = min(self.pull_strenght + 0.2,  0.9999)
#             self.drag_move_set_vert_co(context, event)
#             FRAME_CHANGED = False
#
#         elif event.type == 'LEFTMOUSE' and event.value == 'RELEASE':  # should only be executed on LMB
#             if not self.with_pinning and self.hit_vert_id:
#                 # self.pin_vg.add([self.hit_vert_id], self.previous_vert_weight, 'REPLACE') # XXX: old way - worked in B 3.3
#                 if self.previous_vert_weight > 0.5 and self.hit_vert_id not in self.pinned_verts_ids:
#                     self.pinned_verts_ids.append(self.hit_vert_id)
#             else:
#                 if self.hit_vert_id not in self.pinned_verts_ids:
#                     self.pinned_verts_ids.append(self.hit_vert_id)
#                 self.pin_vg.add([self.hit_vert_id], 1, 'REPLACE')
#             self.with_pinning = False
#             bpy.data.objects.remove(self.empty) # New way
#
#             if context.screen.is_animation_playing and self.payback_method == 'INTERRUPTED':
#                 bpy.ops.screen.animation_play()
#             self.state = 'main'
#
#         elif event.type == 'P' and event.value == 'PRESS':
#             self.with_pinning = not self.with_pinning
#
#         elif event.type == 'U' and event.value == 'PRESS':
#             self.pinned_verts_ids.remove(self.hit_vert_id)
#             self.pin_vg.remove([self.hit_vert_id])
#             self.previous_vert_weight = 0
#             self.with_pinning = False
#
#         elif event.type in {'MIDDLEMOUSE', 'WHEELUPMOUSE', 'WHEELDOWNMOUSE'}:
#             return {'PASS_THROUGH'}
#
#         elif event.type in {'RIGHTMOUSE', 'RET', 'SPACE'} and event.value == 'PRESS':
#             return self.end_finish()
#
#         elif event.type in {'ESC'} and event.value == 'PRESS':
#             return self.end_cancel()
#
#         return {'RUNNING_MODAL'}
#
#
#     def modal(self, context, event):
#         self.event = event
#         context.area.tag_redraw()
#         if self.state == 'main':
#             return self.modal_main(context, event)
#         elif self.state == 'drag_empty':
#             return self.drag_empty(context, event)
#
#         return {'RUNNING_MODAL'}
#         if event.alt and event.type == 'q' and event.value == 'PRESS': #FORCE QUIT
#             return {"CANCELLED"}
#
#
#     def before_start(self, context, event):
#         global FIRST_RUN, DEPSGRAPH, CLOTH_OBJ
#         FIRST_RUN = True
#         self.draw_overlays = False
#         self.state = 'main' # initial state : 'main', 'drag_empty',
#         cloth_obj = context.active_object
#         CLOTH_OBJ = cloth_obj
#         self.cloth_obj = cloth_obj
#         if not cloth_obj.data.shape_keys or not 'rest_shape' in cloth_obj.data.shape_keys.key_blocks.keys():
#             create_rest_shape(context, cloth_obj)
#         rest_shape = cloth_obj.data.shape_keys.key_blocks['rest_shape']
#
#         set_cloth_mod_rest_shape(cloth_obj, rest_shape)
#         self.pin_vg, self.pin_vg_existed = get_pin_vertex_group(cloth_obj)
#         if self.pin_vg_existed:  # do we created pin_vg, or we used existing one from cloth_mod.vertex_group_mass
#             self.pin_vg_weights_backup = get_weights(cloth_obj, self.pin_vg) # for resetting pin_vg to inital weights
#         else:
#             self.pin_vg_weights_backup = []
#
#         self.mw_inv = cloth_obj.matrix_world.inverted() #just for performance
#         self.hit_vert_id = None # id of picked vert with LMB
#         self.with_pinning = False #LMB will pin picked vert permanently
#         self.drag_target_loc = None #helper for gpu drawing only
#         self.pinned_verts_ids = [v_id for v_id, weight in enumerate(self.pin_vg_weights_backup) if weight and weight[0]] #cache for speedup in gpu drawing
#         self.all_vert_ids = [i for i in range(len(cloth_obj.data.vertices))] #another cache for all verts list
#         self.payback_method = context.scene.garment_props.animation_method  # CONSTANT of INTERRUPTED
#         if not bpy.context.screen.is_animation_playing and self.payback_method == 'CONSTANT':
#             bpy.ops.screen.animation_play()
#
#         if frame_change_update not in bpy.app.handlers.frame_change_post:
#             bpy.app.handlers.frame_change_post.append(frame_change_update)
#
#         DEPSGRAPH = context.evaluated_depsgraph_get()
#         return
#
#     def invoke(self, context, event):
#         if context.space_data.type == 'VIEW_3D':
#             if not context.active_object or context.active_object.type != 'MESH' or not any([mod.type == 'CLOTH' for mod in context.active_object.modifiers]):
#                 self.report({'WARNING'}, "Select mesh object first, with cloth modifier. Cancelling")
#                 return {'CANCELLED'}
#             if any([mod.type == 'CLOTH' and mod.point_cache.is_baked for mod in context.active_object.modifiers]):
#                 self.report({'INFO'}, f'Object {context.active_object.name} is cached! Disable clot cache for interaction. Cancelling')
#                 return {'CANCELLED'}
#             self.invoke_basic(context, event)
#             # self.fsm_init()
#             self.before_start(context, event)
#             context.window_manager.modal_handler_add(self)
#             return {'RUNNING_MODAL'}
#         else:
#             self.report({'WARNING'}, "Active space must be a View3d")
#             return {'CANCELLED'}
#
#
