'''
Copyright (C) 2017 JOSECONSCO
Created by JOSECONSCO

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''
import bpy
import numpy as np

def set_cloth_mod_rest_shape(cloth_obj, rest_shape):
    for mod in cloth_obj.modifiers:  # disable constructive modifiers and cloth, for generation of rest shape
        if mod.type == 'CLOTH':
            mod.settings.rest_shape_key = rest_shape

def create_rest_shape(context, cloth_obj):
    if not cloth_obj.data.shape_keys:
        cloth_obj.shape_key_add(name='sim_shape')  # add  base key
    else:
        cloth_obj.data.shape_keys.key_blocks[0].name = 'sim_shape'

    if not 'rest_shape' in cloth_obj.data.shape_keys.key_blocks.keys():
        cloth_obj.shape_key_add(name='rest_shape', from_mix=False)

    rest_shape = cloth_obj.data.shape_keys.key_blocks['rest_shape']
    set_cloth_mod_rest_shape(cloth_obj, rest_shape)

    for mod in cloth_obj.modifiers:  # disable constructive modifiers and cloth, for generation of rest shape
        mod.show_viewport = False

    cloth_obj.show_only_shape_key = False
    cloth_obj.data.shape_keys.use_relative = False

    #write backup of base mesh shape to 'rest_shape'
    verts_co = np.empty((len(cloth_obj.data.vertices)*3), 'f')
    depsgraph = context.evaluated_depsgraph_get()
    depsgraph.objects[cloth_obj.name].data.vertices.foreach_get("co", verts_co)
    rest_shape.data.foreach_set("co", verts_co)

    constructive_modifiers = ['SUBSURF', 'BEVEL', 'MIRROR', 'MASK', 'EDGE_SPLIT', 'SOLIDIFY']
    for mod in cloth_obj.modifiers:  # disable all mods that were baked to  rest_pose (#!maybe not, and leave bend def?)
        mod.show_viewport = True if mod.type not in constructive_modifiers else False


def get_pin_vertex_group(cloth_obj):
    '''Get pin vert group by name or from cloth mod vertex_group_mass property, or create new one '''
    found_pin_vg = None
    pin_existed = False
    for mod in cloth_obj.modifiers:
        if mod.type == 'CLOTH':
            if not mod.settings.vertex_group_mass:
                if 'GT_PIN' in cloth_obj.vertex_groups.keys():
                    pin_existed = True
                else:
                    found_pin_vg = cloth_obj.vertex_groups.new(name='GT_PIN')
                found_pin_vg = cloth_obj.vertex_groups['GT_PIN']
                mod.settings.vertex_group_mass = found_pin_vg.name
            else:
                found_pin_vg = cloth_obj.vertex_groups[mod.settings.vertex_group_mass]
                pin_existed = True
    return found_pin_vg, pin_existed


def inflate():  # ? dosent work. Not sure if it will ever work
    print('INFLaTING')
    cloth_obj = bpy.context.active_object
    vert_count = len(cloth_obj.data.vertices)
    vert_ids = [i for i in range(vert_count)]
    pin_vg, pin_existed = get_pin_vertex_group(cloth_obj)
    pin_vg.add(vert_ids, 1, "ADD")
    depsgraph = bpy.context.evaluated_depsgraph_get()
    cloth_eval = cloth_obj.evaluated_get(depsgraph)  # cloth obj, with modifiers
    vertices = np.empty((vert_count, 3), 'f')
    normal = np.empty((vert_count, 3), 'f')
    cloth_eval.data.vertices.foreach_get("co", np.reshape(vertices, vert_count * 3))
    cloth_eval.data.vertices.foreach_get("normal", np.reshape(normal, vert_count * 3))
    vertices += normal*0.01
    cloth_obj.data.shape_keys.key_blocks[0].data.foreach_set("co", vertices.ravel())  # works without pin, cos first frame?
    cloth_obj.data.shape_keys.key_blocks[0].data.update()
    cloth_obj.data.vertices.foreach_set("co", vertices.ravel())  # works without pin, cos first frame?
    cloth_obj.data.update()
    # pin_vg.remove(vert_ids)
