'''
Copyright (C) 2017 JOSECONSCO
Created by JOSECONSCO

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import bpy
from .utils.helper_functions import get_addon_preferences, fix_garment_vgroups, get_garment, fix_garment_colliding_names
from .cloth_presets_lib import cloth_file_presets


class GTOOL_PT_ClothInteraction(bpy.types.Panel):
    bl_idname = "GTOOL_PT_ClothInteraction"
    bl_label = "Cloth interaction"
    bl_space_type = 'VIEW_3D'
    bl_region_type = 'UI'
    bl_category = 'Garment'
    bl_context = 'objectmode'
    bl_order = 2

    @classmethod
    def poll(cls, context):
        return (context.object is not None)

    def draw(self, context):
        layout = self.layout
        col = layout.column(align=True)

        row = col.row(align=True)
        row.operator('cloth.pull_cloth')
        # row.prop(context.scene.garment_props, 'continuous_sim', text='', icon='FILE_REFRESH')

        # col.prop(context.scene.garment_props, 'animation_method')
        col.prop(context.scene.garment_props, 'pull_force')
        col.prop(context.scene.garment_props, 'pull_falloff')
        row = col.row(align=True)
        row.prop(context.scene.garment_props, 'distance_max')
        ic = 'RADIOBUT_ON' if context.scene.garment_props.use_max_distance else "RADIOBUT_OFF"
        row.prop(context.scene.garment_props, 'use_max_distance', text='', icon=ic)
        # col.operator('cloth.reset_cloth_sim', icon='QUIT')
        # row = layout.column(align=True)
        # row.operator('cloth.create_rest_shape', icon='MOD_CLOTH')

# create UIList with scene.garment_props.local_cloth_presets
class GTOOL_UL_local_cloth_presets(bpy.types.UIList):
    def draw_item(self, context, layout, data, item, icon, active_data, active_propname):
        cloth_preset = item
        row = layout.row()
        split = row.split(factor=0.8, align=True)
        split.prop(cloth_preset, "name", text="", emboss=False, icon_value=icon)
        preset_mat = bpy.data.materials.get(cloth_preset.name)
        if preset_mat:
            split.prop(preset_mat, "diffuse_color", text="")
        else: # run   'cloth.generate_cloth_material' operator
            split.operator('cloth.generate_cloth_material', text='', icon='MATERIAL', emboss=False).preset = cloth_preset.name



class GTOOL_PT_LocalClothPresets(bpy.types.Panel):
    bl_idname = "GTOOL_PT_LocalClothPresets"
    bl_label = "Cloth Physics Presets"
    bl_space_type = 'VIEW_3D'
    bl_region_type = 'UI'
    bl_category = 'Garment'
    # bl_context = 'objectmode'
    bl_order = 1

    @classmethod
    def poll(cls, context):
        return context.active_object

    def draw(self, context):
        layout = self.layout
        garment_props = context.scene.garment_props

        row = layout.row(align=True)
        gprops = context.scene.garment_props
        if len(gprops.local_cloth_presets) > 0:
            active_local_preset = gprops.local_cloth_presets[gprops.active_lpreset_index]
            mat = bpy.data.materials.get(active_local_preset.name)
            op = row.operator("garment.assign_cloth_preset", text='Assign Preset: '+active_local_preset.name, icon_value=layout.icon(mat) if mat else -1)
            op.preset = active_local_preset.name

        active_idx = garment_props.active_lpreset_index
        if len(garment_props.local_cloth_presets) == 0:
            row = layout.row(align=True)
            row.operator('cloth.add_local_preset', icon='ADD')
            row.popover(GTOOL_PT_GlobalClothPresetsPanel.bl_idname, icon='PRESET', text="",)
        else:
            local_cloth_preset = garment_props.local_cloth_presets[active_idx]

            row = layout.row(align=True)
            col = row.column(align=True)
            col.template_list("GTOOL_UL_local_cloth_presets", "", garment_props, "local_cloth_presets", garment_props, "active_lpreset_index", rows=3)
            col = row.column(align=True)
            col.operator('cloth.add_local_preset', icon='ADD', text="")
            col.operator('cloth.delete_local_preset', icon='REMOVE', text="")
            col.popover(GTOOL_PT_GlobalClothPresetsPanel.bl_idname, icon='PRESET', text="",)


            col = layout.column(align=True)
            col.label(text=f"{local_cloth_preset.name} Unique Props")
            col.prop(local_cloth_preset, 'tension_stiffness')
            col.prop(local_cloth_preset, 'compression_stiffness')
            col.prop(local_cloth_preset, 'shear_stiffness')
            col.prop(local_cloth_preset, 'bending_stiffness')

            col = layout.column(align=True)
            col.prop(local_cloth_preset, 'shrink_min')

            col = layout.column(align=True)
            col.label(text="Shared Props")
            col.prop(local_cloth_preset, 'sewing_force')
            col.prop(local_cloth_preset, 'mass')
            col.prop(local_cloth_preset, 'air_damping')

            col = layout.column(align=True)
            col.prop(local_cloth_preset, 'tension_damping')
            col.prop(local_cloth_preset, 'compression_damping')
            col.prop(local_cloth_preset, 'shear_damping')
            col.prop(local_cloth_preset, 'bending_damping')

        active_obj = context.active_object
        if not active_obj:
            return
        cloth_mod = active_obj.modifiers.get('Cloth')
        if cloth_mod:
            col = layout.column(align=True)
            col.label(text="Other settings")
            # cloth_mod.settings.use_sewing_springs = True
            col.prop(cloth_mod.settings, 'quality')
            col.prop(cloth_mod.collision_settings, 'distance_min', text='Colision Distance')
            # use_self_collision - toggle, with .self_distance_min
            row = col.row(align=True)
            ic = 'RADIOBUT_ON' if cloth_mod.collision_settings.use_self_collision else "RADIOBUT_OFF"
            txt = '' if cloth_mod.collision_settings.use_self_collision else 'Enable Self Collision'
            row.prop(cloth_mod.collision_settings, 'use_self_collision', text=txt, icon=ic)
            if cloth_mod.collision_settings.use_self_collision:
                row.prop(cloth_mod.collision_settings, 'self_distance_min', text='Self Collision Distance')

class GTOOL_PT_ClothTools(bpy.types.Panel):
    bl_idname = "GTOOL_PT_ClothTools"
    bl_label = "Pattern (Curve) Tools"
    bl_space_type = 'VIEW_3D'
    bl_region_type = 'UI'
    bl_category = 'Garment'
    # bl_context = 'objectmode'
    bl_order = 4

    @classmethod
    def poll(cls, context):
        return (context.object is not None)

    def draw(self, context):
        cloth_garment_data = context.scene.cloth_garment_data
        layout = self.layout

        row = layout.row(align=True)
        row.operator('curve.curve_symmetrize', icon='ARROW_LEFTRIGHT')
        row.operator('curve.pattern_flip', icon='MOD_MIRROR')
        row = layout.row(align=True)
        row.operator('curve.connect_points')
        row.operator('curve.weld_points')
        row = layout.row(align=True)
        row.operator('curve.separate_pattern_sel')
        row.operator('curve.join_patterns')
        row = layout.row(align=True)
        row.operator('curve.duplicate_sewed', icon='DUPLICATE')
        row.operator('curve.cut_half')
        row = layout.row(align=True)
        row.operator('curve.pattern_cut')



class GT_PresetsSubMenu(bpy.types.Menu):
    bl_idname = "GT_MT_cloth_presets_submenu"
    bl_label = "Select Preset"


    def draw(self, context):
        layout = self.layout

        garment = get_garment(context)
        sewing_pattern = context.active_pattern #if hasattr(context, 'pattern_idx') else -1

        # find pattern idx
        pattern_idx = 0
        for idx, p in enumerate(garment.sewing_patterns):
            if p == sewing_pattern:
                pattern_idx = idx
                break

        # TODO: add default materila to list
        def_mat = bpy.data.materials.get(garment.default_cloth_preset)
        name = '[Default]' if sewing_pattern.cloth_preset == 'Default' else ' Default '
        op = layout.operator("garment.switch_pattern_preset", text=name, icon_value=layout.icon(def_mat) if def_mat else 1)
        op.preset = 'Default'
        op.pattern_index = pattern_idx
        scene = garment.id_data
        gprops = scene.garment_props
        for local_preset in gprops.local_cloth_presets:
            mat = bpy.data.materials.get(local_preset.name)
            name = '['+local_preset.name+']' if sewing_pattern.cloth_preset == local_preset.name else ' '+local_preset.name
            op = layout.operator("garment.switch_pattern_preset", text=name, icon_value=layout.icon(mat) if mat else 1)
            op.preset = local_preset.name
            op.pattern_index = pattern_idx


class GTOOL_PT_GlobalClothPresetsPanel(bpy.types.Panel):
    bl_idname = "GTOOL_PT_GlobalClothPresetsPanel"
    bl_label = "Library Presets"
    bl_space_type = 'VIEW_3D'
    bl_region_type = 'UI'
    bl_context = 'objectmode'

    def draw(self, context):
        layout = self.layout
        # subrow.prop_menu_enum(cloth_preset.preset, 'assigned_patterns')
        layout.label(text="Load Global Preset")
        col = layout.column(align=True)
        for item in cloth_file_presets(self, context):
            row = col.row(align=True)
            op = row.operator('cloth.load_global_cloth_preset', text=item[1])
            op.preset = item[0]
            # row.label(text=item[1])
            row.operator('cloth.delete_global_cloth_preset', icon='REMOVE', text='').preset = item[0]
        col = layout.column(align=True)
        col.operator('cloth.save_global_cloth_preset', icon='ADD')
        col.operator('cloth.reset_global_cloth_presets', icon='FILE_REFRESH')
            # row.operator('cloth.load_preset', text=item[0]).filepath = item[1]


class GT_UL_GarmentsList(bpy.types.UIList):
    def draw_item(self, context, layout, data, item, icon, active_data, active_propname):
        scene = data
        garment = item
        # draw_item must handle the three layout types... Usually 'DEFAULT' and 'COMPACT' can share the same code.
        if self.layout_type in {'DEFAULT', 'COMPACT'}:
            # start your row layout by a label (icon + text), or a non-embossed text field, to make it selectable
            row = layout.row(align=True)
            row.prop(garment, "name", text="", emboss=False)
            show_icon = 'HIDE_OFF' if garment.show_patterns else 'HIDE_ON'
            row.prop(garment, "show_patterns", text="", emboss=False, icon=show_icon)
            row.prop(garment, 'sewings_overlay', text='', emboss=False, icon='LINKED')

        # 'GRID' layout type should be as compact as possible (typically a single icon!).
        elif self.layout_type == 'GRID':
            layout.alignment = 'CENTER'
            row.prop(garment, "name", text="")
            # layout.label(text="", icon_value=icon)


# implement add_item operator for short hair
class HTOOL_OT_CreateProp(bpy.types.Operator):
    bl_idname = 'some_props.add_item'
    bl_label = 'Create Vg To Uv'
    bl_description = 'Create Vg To Uv'
    bl_options = {'REGISTER', 'UNDO'}

    def execute(self, context):
        obj = context.active_object
        some_props = obj.data.some_prop
        some_props.coll_prop_items.add()
        some_props.active_index = len(some_props.coll_prop_items)-1
        return {'FINISHED'}

# implement delete_item operator for short hair
class HTOOL_OT_DeleteProp(bpy.types.Operator):
    bl_idname = 'some_props.delete_item'
    bl_label = 'Delete Vg To Uv'
    bl_description = 'Delete Vg To Uv'
    bl_options = {'REGISTER', 'UNDO'}

    def execute(self, context):
        obj = context.active_object
        some_props = obj.data.some_prop
        some_props.coll_prop_items.remove(some_props.active_index)
        cnt = len(some_props.coll_prop_items)
        some_props.active_index = some_props.active_index if some_props.active_index < cnt else cnt-1
        return {'FINISHED'}

class GT_RightSiderPanel:
    bl_space_type = 'VIEW_3D'
    bl_region_type = 'UI'
    bl_category = 'Garment'
    bl_context = 'objectmode'
    # bl_options = {'DEFAULT_CLOSED'}

class GT_PT_MainGarmentsPanel(GT_RightSiderPanel, bpy.types.Panel):
    bl_label = "Garments"
    bl_idname = "GT_PT_MainGarmentsPanel"

    # @classmethod
    # def poll(cls, context):
    #     return context.active_object

    def draw(self, context):
        layout = self.layout
        col = layout.column(align = True)
        row = col.row(align=True)
        row.template_list("GT_UL_GarmentsList", "", context.scene, "cloth_garment_data", context.scene, "garment_index", rows=2)
        col = row.column(align=True)
        col.operator('garment.add_garment', text='', icon='ADD')
        col.operator('garment.remove_garment', icon='REMOVE', text='')

class GTOOL_PT_ClothPanel(GT_RightSiderPanel, bpy.types.Panel):
    bl_idname = "GTOOL_PT_ClothPanel"
    bl_label = "Edit / Generate"
    bl_parent_id = "GT_PT_MainGarmentsPanel"
    bl_order = 5

    def draw(self, context):
        layout = self.layout
        if len(context.scene.cloth_garment_data)==0:
            row = layout.row(align=True)
            row.operator('garment.add_garment', text='Create Garment', icon='ADD')

        garment = get_garment(context)
        if not garment:
            return

        col = layout.column(align = True)
        row = col.row(align=True)
        # row.prop(garment, 'sewings_overlay', icon='LINKED', text='')
        row.operator('cloth.raycast_edit_sewing', text=f'Edit Sewings ({len(garment.garment_sewings)})')
        row = col.row(align=True)
        row.operator('cloth.edit_bend', icon='MOD_SIMPLEDEFORM') # apply this automatically (smooth = 10)
        row.operator('garment.update_bend_node_group', icon="FILE_REFRESH", text='')  # apply this automatically (smooth = 10)
        row = col.row(align=True)
        row.operator('garment.raycast_edit_pocket', icon='LINKED', text=f'Edit Pockets ({len(garment.pockets)})')
        on_ico = 'RADIOBUT_ON' if garment.enable_pockets else "RADIOBUT_OFF"
        row.prop(garment, 'enable_pockets', icon=on_ico, text='')
        row = col.row(align=True)
        row.operator('garment.raycast_edit_pins', icon='LINKED', text=f'Edit Pins ({len(garment.pins)})')
        on_ico = 'RADIOBUT_ON' if garment.enable_pins else "RADIOBUT_OFF"
        row.prop(garment, 'enable_pins', icon=on_ico, text='')

        col = layout.column(align=True)
        row = col.row(align=True)
        split = row.split(factor=0.65, align=True)
        warn_ico = 'ERROR' if garment.predicted_out_vert_count > 15000 else 'NONE'
        split.operator('garment.garment_to_mesh', icon=warn_ico)
        split.prop(garment, 'output_topology', text='', icon='MOD_DATA_TRANSFER')  # apply this automatically (smooth = 10)
        row = col.row(align=True)
        row.prop(garment, 'cloth_res', text='Resolution')
        row = col.row(align=True)
        row.prop(garment, 'target_cloth_object')  # apply this automatically (smooth = 10)

        col = layout.column(align = True)
        col.operator('garment.init_simulation')
        row = col.row(align=True)
        split = row.split(factor=0.7, align=True)
        split.prop(garment, 'sim_time')
        split.prop(garment, 'warmup_factor', text='')


        row = col.row(align=True)
        cache = None
        if garment.target_cloth_object and garment.target_cloth_object.modifiers.get('Cloth'):
            mod = garment.target_cloth_object.modifiers['Cloth']
            cache = mod.point_cache
            row.context_pointer_set("point_cache", cache)
            # row.operator_context = 'INVOKE_DEFAULT'
        if cache and cache.is_baked is True:
            row.operator("ptcache.free_bake", text="Delete Bake")
        else:
            row.operator("ptcache.bake", text="Bake").bake = True
        # row.operator('garment.simulate_cloth')
        col = layout.column(align=True)
        col.operator('garment.finalize_cloth_mesh') # apply this automatically (smooth = 10)
        col.operator('cloth.save_to_shape')

        prefs = get_addon_preferences()
        if prefs.debug_view:
            row = layout.row(align=True)
            row.prop(garment, 'merge_patterns', expand=True)



class GTOOL_PT_PatternsPanel(GT_RightSiderPanel, bpy.types.Panel):
    bl_idname = "GTOOL_PT_PatternsPanel"
    bl_label = "Patterns"
    bl_options = {'DEFAULT_CLOSED'}
    bl_parent_id = "GT_PT_MainGarmentsPanel"
    bl_order = 10

    def draw(self, context):
        layout = self.layout

        garment = get_garment(context)
        if not garment:
            return

        col = layout.column(align=True)
        if len(garment.sewing_patterns) > 0:
            mat = bpy.data.materials.get(garment.default_cloth_preset)
            col.menu("GARMENT_MT_SelDefaultPreset", text=f'Default Preset: {garment.default_cloth_preset}', icon_value=layout.icon(mat) if mat else -1)

            col = layout.column(align=True)
            for pattern_idx, sewing_pattern in enumerate(garment.sewing_patterns):
                row2 = col.row(align=True)
                enabled = 'RADIOBUT_ON' if sewing_pattern.is_enabled else "RADIOBUT_OFF"
                row2.prop(sewing_pattern, 'is_enabled', text='', emboss=True, icon=enabled)

                ic = 'TRIA_RIGHT' if context.active_object and context.active_object == sewing_pattern.pattern_obj else 'MATCLOTH'
                # row2.label(text=sewing_pattern.pattern_obj.name, icon=ic)
                row2.prop(sewing_pattern, "pattern_obj", text='', icon=ic)
                row2.prop(sewing_pattern, 'add_lattice', text='', icon='MOD_LATTICE')
                # row2.prop(sewing_pattern, 'add_bend_deform', text='', icon='MOD_SIMPLEDEFORM')
                name = garment.default_cloth_preset if sewing_pattern.cloth_preset == 'Default' else sewing_pattern.cloth_preset
                mat = bpy.data.materials.get(name)
                # row2.prop(sewing_pattern, 'cloth_preset', text='', icon_value=layout.icon(mat) if mat else -1, icon_only=True) #  icon='PRESET'

                row2.context_pointer_set("active_garment", garment)  # done: can wa pass value to Menu?? With This workaround yes
                row2.context_pointer_set("active_pattern", sewing_pattern)  # done: can wa pass value to Menu?? With This workaround yes
                row2.menu(menu="GT_MT_cloth_presets_submenu", text="", icon_value=layout.icon(mat) if mat else 1)
                # row.prop_menu_enum(garment, 'default_cloth_preset', text='', icon='PRESET')

                oper = row2.operator('garment.pattern_remove', icon='REMOVE', text='')  # ! replaced by X on pattern_obj property. To fix gen_vgroups set breakage
                oper.pattern_idx = pattern_idx

        col.operator('garment.add_pattern', icon='ADD')

        prefs = get_addon_preferences()
        if prefs.debug_view: #* Sewings
            box2 = layout.box()
            row = box2.row(align=True)
            row.label(text='Sewings:')
            if len(garment.garment_sewings) > 0:
                for sewing_index, garment_sewing in enumerate(garment.garment_sewings):
                    row_main = box2.row(align=True)
                    sub_col = row_main.column(align=True)
                    subrow = sub_col.row(align=True)
                    subrow.prop(garment_sewing, "source_obj",  text='')
                    subrow.prop(garment_sewing, "from_spline_idx")
                    subrow.prop(garment_sewing, "from_segment_idx")

                    subrow = sub_col.row(align=True)
                    subrow.prop(garment_sewing, "target_obj", text='')
                    subrow.prop(garment_sewing, "to_spline_idx")
                    subrow.prop(garment_sewing, "to_segment_idx")

                    sub_col = row_main.column(align=True)
                    subrow = sub_col.row(align=True)
                    subrow.prop(garment_sewing, "flip", text='', icon='CENTER_ONLY')
                    subrow = sub_col.row(align=True)
                    pair = subrow.operator('sewing.sewing_remove', icon='REMOVE', text='')
                    pair.sewing_index = sewing_index



class GTOOL_PT_VertGroupsPanel(GT_RightSiderPanel, bpy.types.Panel):
    bl_idname = "GTOOL_PT_VertGroupsPanel"
    bl_label = "Vertex Groups"
    bl_parent_id = "GT_PT_MainGarmentsPanel"
    bl_order = 20

    def draw(self, context):
        layout = self.layout

        garment = get_garment(context)
        if not garment:
            return

        row_main = layout.row(align=True)
        sub_col = row_main.column(align=True)
        for vg_idx, vg in enumerate(garment.generated_vgroups):
            subrow = sub_col.row(align=True)
            subrow.prop(vg, 'name', text='')
            subrow.prop_menu_enum(vg, 'vgroups_patterns')
            # subrow.prop(vg, 'add_bend_deform', text='', icon='MOD_SIMPLEDEFORM')
            subrow.prop(vg, 'add_lattice', text='', icon='MOD_LATTICE')
            # row.prop(vg, 'add_bend_deform', text='', icon='MOD_SIMPLEDEFORM')
            pair = subrow.operator('garment.vg_remove', icon='REMOVE', text='')
            pair.vg_idx = vg_idx
        sub_col.operator('garment.vg_add', icon='ADD')

        row = layout.row(align=True)
        row.alignment = 'EXPAND'
        # row.prop(garment, 'add_border_mask', icon='FULLSCREEN_ENTER', emboss=True)
        # row.prop(garment, 'add_non_border_mask', icon='FULLSCREEN_EXIT', emboss=True)

        prefs = get_addon_preferences()
        if prefs.debug_view:  # * Pockets
            box2 = layout.box()
            box2.label(text='Pockets')
            if len(garment.pockets) > 0:
                for pocket_idx, garment_pocket in enumerate(garment.pockets):
                    col = box2.column(align=True)
                    row = col.row(align=True)
                    row.prop(garment_pocket, "pattern_obj", text='Pocket Obj')
                    row.prop(garment_pocket, "target_pattern")
                    row.prop_menu_enum(garment_pocket, "pocket_sewing")
                    oper = row.operator('garment.pocket_remove', icon='REMOVE', text='')
                    oper.pocket_idx = pocket_idx

        if prefs.debug_view:  # * Pins
            box2 = layout.box()
            box2.label(text='Pins')
            if len(garment.pins) > 0:
                for pin_idx, garment_pin in enumerate(garment.pins):
                    box3 = box2.box()
                    col = box3.column(align=True)
                    row = col.row(align=True)
                    row.prop(garment_pin, "source_obj")
                    row.prop(garment_pin, "source_co")
                    row = col.row(align=True)
                    row.prop(garment_pin, "target_obj")
                    row.prop(garment_pin, "target_co")
                    oper = row.operator('garment.pin_remove', icon='REMOVE', text='')
                    oper.pin_idx = pin_idx


# create menu with default_cloth_preset enum properties list
class GARMENT_MT_SelDefaultPreset(bpy.types.Menu):
    bl_idname = "GARMENT_MT_SelDefaultPreset"
    bl_label = "Default Preset"
    bl_description = "Select Default Cloth Preset, that will be used on patterns that use 'Default' preset"

    def draw(self, context):
        layout = self.layout
        gprops = context.scene.garment_props
        row = layout.row(align=True)
        mat = bpy.data.materials.get('None')
        op = row.operator("garment.assign_default_preset", text = 'None', icon_value=layout.icon(mat) if mat else -1)
        op.preset = 'None'
        for lpreset in gprops.local_cloth_presets:
            row = layout.row(align=True)
            mat = bpy.data.materials.get(lpreset.name)
            op = row.operator("garment.assign_default_preset", text = f'{lpreset.name}', icon_value=layout.icon(mat) if mat else -1)
            op.preset = lpreset.name


class GTOOL_OT_GarmentAdd(bpy.types.Operator):
    bl_idname = "garment.add_garment"
    bl_label = "Add Garment"
    bl_description = 'Add new Garment. You will be able to assign sewing patterns to it, link them with sewings and finally generate triangulated cloth mesh.'
    bl_options = {'REGISTER', 'UNDO'}

    def execute(self, context):
        cloth_garment_data = context.scene.cloth_garment_data
        new_garment = cloth_garment_data.add()
        fix_garment_colliding_names(new_garment, context)
        context.scene.garment_index = len(cloth_garment_data) - 1
        return {'FINISHED'}


class GTOOL_OT_GarmentRemove(bpy.types.Operator):
    bl_idname = "garment.remove_garment"
    bl_label = "Remove Garment"
    bl_description = "Remove Garment"
    bl_options = {'REGISTER', 'UNDO'}


    def execute(self, context):
        cloth_garment_data = context.scene.cloth_garment_data
        garment_index = context.scene.garment_index
        garment = cloth_garment_data[garment_index]
        for pattern in garment.sewing_patterns:
            if pattern.pattern_obj and pattern.pattern_obj.name not in context.scene.objects:
                bpy.data.objects.remove(pattern.pattern_obj)
        if garment.target_cloth_object and garment.target_cloth_object.name not in context.scene.objects:
            bpy.data.objects.remove(garment.target_cloth_object)
        cloth_garment_data.remove(garment_index)
        context.scene.garment_index = max(0, garment_index -1)
        return {'FINISHED'}


class GTOOL_OT_SewingPatterPairAdd(bpy.types.Operator):
    bl_idname = "garment.add_pattern"
    bl_label = "Add Pattern"
    bl_description = 'Add Pattern curve to garment'
    bl_options = {'REGISTER', 'UNDO'}

    def execute(self, context):
        garment_index = context.scene.garment_index
        cloth_garment_data = context.scene.cloth_garment_data[garment_index]
        cloth_garment_data.sewing_patterns.add()
        return {'FINISHED'}


class GTOOL_OT_GarmetnPocketAdd(bpy.types.Operator):
    bl_idname = "garment.pocket_add"
    bl_label = "Add Pocket"
    bl_description = 'Add Pocket curve to garment'
    bl_options = {'REGISTER', 'UNDO'}

    garment_index: bpy.props.IntProperty()

    def execute(self, context):
        cloth_garment_data = context.scene.cloth_garment_data[self.garment_index]
        cloth_garment_data.pockets.add()
        return {'FINISHED'}


class GTOOL_OT_GarmetnPinAdd(bpy.types.Operator):
    bl_idname = "garment.pin_add"
    bl_label = "Add Pin"
    bl_description = 'Add Pin curve to garment'
    bl_options = {'REGISTER', 'UNDO'}

    garment_index: bpy.props.IntProperty()

    def execute(self, context):
        cloth_garment_data = context.scene.cloth_garment_data[self.garment_index]
        cloth_garment_data.pins.add()
        return {'FINISHED'}

class GTOOL_OT_VGAdd(bpy.types.Operator):
    bl_idname = "garment.vg_add"
    bl_label = "Add vertex group"
    bl_description = 'Generate new vertex group and assign patterns to it'
    bl_options = {'REGISTER', 'UNDO'}

    def execute(self, context):
        garment_index = context.scene.garment_index
        cloth_garment_data = context.scene.cloth_garment_data[garment_index]
        cloth_garment_data.generated_vgroups.add()
        return {'FINISHED'}


class GTOOL_OT_VGRemove(bpy.types.Operator):
    bl_idname = "garment.vg_remove"
    bl_label = "Remove Vertex Group"
    bl_description = 'Remove merged vertex group'
    bl_options = {'REGISTER', 'UNDO'}

    vg_idx: bpy.props.IntProperty()

    def execute(self, context):
        garment_index = context.scene.garment_index
        cloth_garment_data = context.scene.cloth_garment_data[garment_index]
        cloth_garment_data.generated_vgroups.remove(self.vg_idx)
        return {'FINISHED'}


class GTOOL_OT_ClothPresetRemove(bpy.types.Operator):
    bl_idname = "cloth.cloth_preset_remove"
    bl_label = "Remove cloth preset"
    bl_description = 'Remove cloth preset'
    bl_options = {'REGISTER', 'UNDO'}

    garment_index: bpy.props.IntProperty()
    preset_idx: bpy.props.IntProperty()

    def execute(self, context):
        cloth_garment_data = context.scene.cloth_garment_data[self.garment_index]
        cloth_garment_data.cloth_materials.remove(self.preset_idx)
        return {'FINISHED'}


class GTOOL_OT_SewingPatterPairRemove(bpy.types.Operator):
    bl_idname = "garment.pattern_remove"
    bl_label = "Remove Pattern"
    bl_description = 'Remove Pattern curve from garment (also removes sewing assigned to this Patternt)'
    bl_options = {'REGISTER', 'UNDO'}

    pattern_idx: bpy.props.IntProperty()

    def execute(self, context):
        garment_index = context.scene.garment_index
        garment_settings = context.scene.cloth_garment_data[garment_index]
        garment_mesh = garment_settings.sewing_patterns[self.pattern_idx].pattern_obj
        sewing_to_remove = []
        for i,sewing_data in enumerate(garment_settings.garment_sewings): #remove sewing that are using removed mesh pattern
            if sewing_data.source_obj == garment_mesh or sewing_data.target_obj == garment_mesh:
                sewing_to_remove.append(i)
        for sewing_id in reversed(sewing_to_remove):
            garment_settings.garment_sewings.remove(sewing_id)

        fix_garment_vgroups(garment_settings, removed_id=self.pattern_idx)
        garment_settings.sewing_patterns.remove(self.pattern_idx)

        return {'FINISHED'}

class GTOOL_OT_GarmentPocketRemove(bpy.types.Operator):
    bl_idname = "garment.pocket_remove"
    bl_label = "Remove Pocket"
    bl_description = 'Remove Pocket curve from garment (also removes sewing assigned to this Patternt)'
    bl_options = {'REGISTER', 'UNDO'}

    pocket_idx: bpy.props.IntProperty()

    def execute(self, context):
        garment_index = context.scene.garment_index
        garment_settings = context.scene.cloth_garment_data[garment_index]
        garment_settings.pockets.remove(self.pocket_idx)
        return {'FINISHED'}


class GTOOL_OT_GarmentPinRemove(bpy.types.Operator):
    bl_idname = "garment.pin_remove"
    bl_label = "Remove pin"
    bl_description = 'Remove pin from garment'
    bl_options = {'REGISTER', 'UNDO'}

    pin_idx: bpy.props.IntProperty()

    def execute(self, context):
        garment_index = context.scene.garment_index
        garment_settings = context.scene.cloth_garment_data[garment_index]
        garment_settings.pins.remove(self.pin_idx)
        return {'FINISHED'}


class GTOOL_OT_SewingAdd(bpy.types.Operator):
    bl_idname = "sewing.sewing_add"
    bl_label = "Add Sewing"
    bl_description = 'Add Sewing to garment'
    bl_options = {'REGISTER', 'UNDO'}

    do_modal: bpy.props.BoolProperty(default=False)

    def execute(self, context):
        garment = get_garment(context)
        garment.garment_sewings.add()
        if self.do_modal:
            bpy.ops.cloth.raycast_edit_sewing('INVOKE_DEFAULT', sewing_index=len(garment.garment_sewings)-1)
        return {'FINISHED'}


class GTOOL_OT_SewingRemove(bpy.types.Operator):
    bl_idname = "sewing.sewing_remove"
    bl_label = "Remove Sewing"
    bl_description = 'Remove Sewing from garment'
    bl_options = {'REGISTER', 'UNDO'}

    sewing_index: bpy.props.IntProperty()

    def execute(self, context):
        garment_index = context.scene.garment_index
        context.scene.cloth_garment_data[garment_index].garment_sewings.remove(self.sewing_index)
        return {'FINISHED'}



class GTOOL_OT_GarmentObjectPicker(bpy.types.Operator):
    bl_idname = "curve.curve_picker"
    bl_label = "Use Selected Object"
    bl_description = 'Use Selected Object'
    bl_options = {'REGISTER', 'UNDO'}

    garment_index: bpy.props.IntProperty()
    pattern_idx: bpy.props.IntProperty()

    def execute(self, context):
        if context.active_object and context.active_object.select_get() and context.active_object.type == 'CURVE':
            context.scene.cloth_garment_data[self.garment_index].sewing_patterns[self.pattern_idx].pattern_obj = context.active_object
        return {'FINISHED'}



class GTOOL_OT_PatternRemoveSewings(bpy.types.Operator):
    bl_idname = "garment.remove_pattern_sewings"
    bl_label = "Remove sewings from pattern"
    bl_description = 'Remove sewings attached to current pattern'
    bl_options = {'REGISTER', 'UNDO'}


    pattern_idx: bpy.props.IntProperty()

    def execute(self, context):
        garment = get_garment(context)

        pattern_obj = garment.sewing_patterns[self.pattern_idx].pattern_obj
        sewing_to_remove = []
        for i,sewing_data in enumerate(garment.garment_sewings): #remove sewing that are using removed mesh pattern
            if sewing_data.source_obj == pattern_obj or sewing_data.target_obj == pattern_obj:
                sewing_to_remove.append(i)
        for sewing_id in reversed(sewing_to_remove):
            garment.garment_sewings.remove(sewing_id)
        return {'FINISHED'}


class GTOOL_OT_AssignDefultPreset(bpy.types.Operator):
    bl_idname = "garment.assign_default_preset"
    bl_label = "Assign Default Preset"
    bl_description = 'Assign Default Preset'
    bl_options = {'REGISTER', 'UNDO'}

    preset: bpy.props.StringProperty()

    def execute(self, context):
        garment = get_garment(context)
        garment.default_cloth_preset = self.preset
        return {'FINISHED'}
