'''
Copyright (C) 2017 JOSECONSCO
Created by JOSECONSCO

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''
import gpu
import bpy
import gpu
import bmesh
from gpu_extras.batch import batch_for_shader
from mathutils import Vector
from mathutils.bvhtree import BVHTree
from bpy_extras import view3d_utils
from .utils.fsm_oper import ModalCommon, CookieCutter_FSM
from .utils.helper_functions import draw_text_line


class GTOOL_OT_RayCastEditPockets(bpy.types.Operator, ModalCommon, CookieCutter_FSM):
    """Modal object selection with a ray cast"""
    bl_idname = "garment.raycast_edit_pocket"
    bl_label = "Edit Pockets"
    bl_description = "Pockets - 2D curves that will be projected and sewed to target pattern"
    bl_options = {'REGISTER', 'UNDO'}

    pocket_index: bpy.props.IntProperty(default=0)
    pocket_sewing_idx = 1

    def draw_text(self, cla, context):
        garment = context.scene.cloth_garment_data[self.garment_index]
        pockets_count = len(garment.pockets)
        draw_text_lines = []
        # draw_text_lines.append(["State: "+self._state,'H1'])
        if self._state == 'main':
            draw_text_lines.append(["[LMB] Select/Add pocket pattern", 'p'])
            draw_text_lines.append(["[DEL/R] Remove Sewing", 'p'])
            draw_text_lines.append(["[ENTER] Confirm", 'p'])
            draw_text_lines.append(["[ESC/RMB] Cancel", 'p'])
        if self._state == 'pick_pocket_target':
            draw_text_lines.append(["[LMB] Select target obj", 'H1'])
            draw_text_lines.append(["[ESC/RMB] Cancel", 'p'])
        if self._state == 'manage_pocket_sewing':
            draw_text_lines.append(["[LMB] Enable/Disable sewing", 'p'])
            draw_text_lines.append(["[DEL/R] Remove Pocket", 'p'])
            draw_text_lines.append(["[ENTER] Confirm", 'p'])
            # draw_text_lines.append(["Editing sewing: " + str(self.pocket_sewing_idx+1) + "/" + str(len(self.patter_segments_cache[self.pocket_obj _name])), 'p'])
        # draw_text_lines.append(["Editing Pocket: " + str(self.pocket_index+1) + "/" + str(pockets_count), 'p'])
        draw_text_line(draw_text_lines)


    def draw_px(self, cla, context):
        if len(self.garment.pockets) == 0:
            return
        # if self._state in ['main', 'manage_pocket_sewing']:  # hover highlight
        shader=self.shader
        gpu.state.blend_set("ALPHA")
        gpu.state.point_size_set(6)
        gpu.state.line_width_set(1.0)
        gpu.state.depth_test_set("LESS")
        for p_id in self.get_visible_pockets():
            current_pocket = self.garment.pockets[p_id]
            if current_pocket.pattern_obj and current_pocket.target_pattern:
                pocket_segments = self.patter_segments_cache[current_pocket.pattern_obj.name][0]
                target_segments = self.patter_segments_cache[current_pocket.pattern_obj.name+'target']

                pocket_active_sewing = [int(x) for x in current_pocket.pocket_sewing]
                for seg_id, (source, target) in enumerate(zip(pocket_segments, target_segments)):
                    if seg_id in pocket_active_sewing:
                        draw_points = []
                        for (point_from, point_to) in zip(source, target):
                            draw_points.extend([(point_from.x, point_from.y, point_from.z), (point_to.x, point_to.y, point_to.z)])
                            col = (0.2, 0.9, 0.2, 0.5) if p_id == self.pocket_index and seg_id == self.pocket_sewing_idx else (1.0, 0.8, 0.0, 0.7)
                            batch = batch_for_shader(shader, 'LINES', {"pos": draw_points})
                            shader.bind()  # draw currently defined sewing
                            shader.uniform_float("color", col)
                            batch.draw(shader)

                        if p_id == self.pocket_index and self._state == 'manage_pocket_sewing':
                            active_segment = []
                            for seg_id, segment in enumerate(pocket_segments):
                                if self.pocket_sewing_idx == seg_id:
                                    active_segment.extend([(point[0], point[1], point[2]) for point in segment])

                            batch = batch_for_shader(shader, "POINTS", {"pos": active_segment})
                            shader.bind()
                            shader.uniform_float("color", (1, 1, 1, 0.8))
                            batch.draw(shader)
        #restore opengl defaults
        gpu.state.depth_test_set("NONE")
        gpu.state.line_width_set(1)
        gpu.state.point_size_set(1)
        gpu.state.blend_set("NONE")

    def get_visible_pockets(self):
        return [idx for idx, pocket in enumerate(self.garment.pockets) if pocket.pattern_obj in self.context.visible_objects and pocket.target_pattern in self.context.visible_objects]

    def cache_target_points(self, pocket_index):
        current_pocket = self.garment.pockets[pocket_index]
        if current_pocket.target_pattern and current_pocket.pattern_obj:
            #get verts from pockets
            target_mat_inv = current_pocket.target_pattern.matrix_world.inverted()
            self.patter_segments_cache[current_pocket.pattern_obj.name+'target'] = [] #cos we store projected points for pocket_obj
            for segment_points in self.patter_segments_cache[current_pocket.pattern_obj.name][0]:
                #if current sewing and prevous are neibors then skip first vert\
                # segment_points[0] - use only first spline for projection???
                target_verts_co = [target_mat_inv@ point for point in segment_points]  # points in cache are already in world space m_w @ p.co
                flat_projected_points = [current_pocket.target_pattern.matrix_world@Vector((p.x, p.y, 0)) for p in target_verts_co]
                self.patter_segments_cache[current_pocket.pattern_obj.name+'target'].append(flat_projected_points)

    def pocket_idx_set(self):
        if self.sorted_hover_sewings:
            self.pocket_index, self.pocket_sewing_idx = self.sorted_hover_sewings[self.sewing_id_offset % len(self.sorted_hover_sewings)]
        else:
            self.pocket_index = -1
            self.pocket_sewing_idx = -1
            self.sewing_id_offset = 0


    def get_curve_pocket(self, obj):
        for p_id, pocket in enumerate(self.garment.pockets):
            if pocket.pattern_obj == obj:
                return p_id
        return -1

    def find_closest_pocket_sewing(self):
        ''' Return index of sewing closest to mouse (form bvh meshes made from sewings)'''
        region = self.context.region
        rv3d = self.context.region_data
        scene = self.context.scene
        region = self.context.region
        rv3d = self.context.region_data
        coord = self.event.mouse_region_x, self.event.mouse_region_y

        # get the ray from the viewport and mouse
        view_vector = view3d_utils.region_2d_to_vector_3d(region, rv3d, coord)
        ray_origin = view3d_utils.region_2d_to_origin_3d(region, rv3d, coord)

        # cast rays and find the closest object
        visible_pockets_ids = self.get_visible_pockets()
        distances = []
        hovered_sewings = []
        for p_idx in visible_pockets_ids:
            pocket_sewings_bvhTrees = self.pocket_sewings_bvht[p_idx]
            for p_sew_id, p_sewing_bvh in enumerate(pocket_sewings_bvhTrees):
                if p_sewing_bvh:
                    loc, normal, face_id, dist = p_sewing_bvh.ray_cast(ray_origin, view_vector)
                    if loc:
                        distances.append(dist)
                        hovered_sewings.append((p_idx, p_sew_id))  # self.garment.garment_sewings

        if len(hovered_sewings) != len(self.sorted_hover_sewings):  # if seg count under cursor is different...
            self.sewing_id_offset = 0
        return [s_id for _, s_id in sorted(zip(distances, hovered_sewings))]  # sorted_hover_sewings by dist to cam. What about artho


    @CookieCutter_FSM.FSM_State('main')
    def modal_main(self):
        if self.event.type == 'LEFTMOUSE' and self.event.value == 'PRESS':
            hit_c, loc_w, curve_obj = self.find_curve_under_cursor(self.event)
            if hit_c:  # picking pocket obj
                pattern_id = self.get_pattern_from_curve(curve_obj)
                if pattern_id >= 0: # can't make pattern from pattern objs.
                    self.report({'WARNING'}, f'Can\'t use sewing pattern as pocket! Remove it from sewing patterns first')
                    return
                bpy.ops.object.select_all(action='DESELECT')
                curve_obj.select_set(True)
                clicked_pocked_id = self.get_curve_pocket(curve_obj)
                if clicked_pocked_id >= 0:
                    self.pocket_index = clicked_pocked_id
                    return 'manage_pocket_sewing'
                else:
                    new_pocket = self.garment.pockets.add()
                    self.pocket_index = len(self.garment.pockets) - 1
                    new_pocket.pattern_obj = curve_obj
                    enabled_sewing = [str(i) for i in range(len(curve_obj.data.splines[0].bezier_points))]
                    new_pocket.pocket_sewing = set(enabled_sewing)
                    return 'pick_pocket_target'
            else:
                self.report({'INFO'}, 'No close pattern detected. Click Again')
                return # stay in this state

        elif self.event.type == 'SPACE' and self.event.value == 'PRESS':
            return 'manage_pocket_sewing'

        elif self.event.type == 'MOUSEMOVE':
            self.sorted_hover_sewings = self.find_closest_pocket_sewing()
            self.pocket_idx_set()

        elif self.event.type == 'A' and self.event.value == 'PRESS':
            self.garment.pockets.add()
            self.pocket_index = len(self.garment.pockets) - 1

        elif (self.event.type == 'R' or self.event.type == 'DEL') and self.event.value == 'PRESS':
            if self.pocket_index != -1:  # do not allow to delete last sewing  so len(sewing) >= 2
                highlighted_pocket = self.garment.pockets[self.pocket_index]
                if str(self.pocket_sewing_idx) in highlighted_pocket.pocket_sewing:
                    highlighted_pocket.pocket_sewing = highlighted_pocket.pocket_sewing.difference(str(self.pocket_sewing_idx))
                self.deselect_all()
                self.pocket_sewings_bvht = self.update_pockets_bvht()
                self.pocket_index = -1

        elif self.event.type in {'MIDDLEMOUSE', 'WHEELUPMOUSE', 'WHEELDOWNMOUSE'}:
            self.allow_nav = True

        elif self.event.type in {'RIGHTMOUSE', 'ESC'} and self.event.value == 'PRESS':
            self.end_cancel()
            return

        elif self.event.type == 'RET' and self.event.value == 'PRESS':
            self.end_commit()
            return 'main'


    @CookieCutter_FSM.FSM_State('pick_pocket_target')
    def pick_pocket_target(self):
        if self.event.type == 'LEFTMOUSE' and self.event.value == 'PRESS':
            hit_c, loc_w, curve_obj = self.find_curve_under_cursor(self.event)
            if hit_c:
                current_pocket = self.garment.pockets[self.pocket_index]
                if curve_obj.name == current_pocket.pattern_obj.name:
                    self.report({'INFO'}, 'Target Cant be the same as pocket object. Click Again')
                    return  # stay in this state
                current_pocket.target_pattern = curve_obj
                self.assign_target_obj_to_garment(curve_obj)
                self.cache_target_points(self.pocket_index)
                self.report({'INFO'}, f'Pocket Added')
                return 'manage_pocket_sewing'
            else:
                self.report({'INFO'}, 'No close pattern detected. Click Again')
                return  # stay in this state

        elif self.event.type in {'MIDDLEMOUSE', 'WHEELUPMOUSE', 'WHEELDOWNMOUSE'}:
            self.allow_nav = True

        elif self.event.type in {'RIGHTMOUSE', 'ESC'} and self.event.value == 'PRESS':
            self.garment.pockets.remove(self.pocket_index)
            self.deselect_all()
            self.pocket_index = -1
            return 'main'


    @CookieCutter_FSM.FSM_State('manage_pocket_sewing')
    def manage_pocket_sewing(self): #disable enable sewing, cycle through them
        current_pocket = self.garment.pockets[self.pocket_index]
        pocket_sewing_count = len(self.patter_segments_cache[current_pocket.pattern_obj.name][0]) if current_pocket is not None else 0
        if self.event.type == 'LEFTMOUSE' and self.event.value == 'PRESS':
            if str(self.pocket_sewing_idx) in current_pocket.pocket_sewing:
                current_pocket.pocket_sewing = current_pocket.pocket_sewing.difference(str(self.pocket_sewing_idx))
            else:
                current_pocket.pocket_sewing = current_pocket.pocket_sewing.union(str(self.pocket_sewing_idx))
            self.pocket_sewings_bvht = self.update_pockets_bvht()

        elif self.event.type == 'MOUSEMOVE':
            # self.sorted_hover_sewings = self.find_closest_pocket_sewing()
            spl_id, self.pocket_sewing_idx = self.find_closest_segment(current_pocket.pattern_obj.name)
            # self.pocket_idx_set_only_sel_pocket()  # update self.pocket_sewing_idx

        elif self.event.type in {'MIDDLEMOUSE', 'WHEELUPMOUSE', 'WHEELDOWNMOUSE'}:
            self.allow_nav = True

        elif (self.event.type == 'R' or self.event.type == 'DEL') and self.event.value == 'PRESS':
            self.garment.pockets.remove(self.pocket_index)
            self.deselect_all()
            self.report({'INFO'}, f'Removed Pocket')
            return 'main'

        elif (self.event.type == 'RET' or self.event.type == 'SPACE') and self.event.value == 'PRESS':
            self.deselect_all()
            return 'main'

    def update_pockets_bvht(self):
        # resampled_splines_segments = self.patter_segments_cache[obj.name]
        out_bvht = []  # 2d array - out_bvht[pocket_id][sewing_id] = bvhtree
        for p_idx, pocket in enumerate(self.garment.pockets):
            out_bvht.append([])
            if pocket.pattern_obj and pocket.target_pattern:
                if pocket.pattern_obj.name not in self.patter_segments_cache or pocket.target_pattern.name not in self.patter_segments_cache:
                    print(f'no cache for {pocket.pattern_obj.name}')
                    print(f'{self.patter_segments_cache}')
                    continue
                pocket_segments = self.patter_segments_cache[pocket.pattern_obj.name][0]
                target_segments = self.patter_segments_cache[pocket.pattern_obj.name+'target']

                pocket_active_sewings = [int(x) for x in pocket.pocket_sewing]
                #* add new bvhtree for each active sewing
                for seg_id, (source_pts, target_pts) in enumerate(zip(pocket_segments, target_segments)):
                    if seg_id in pocket_active_sewings:
                        bm = bmesh.new()   # create an empty BMesh
                        prev_from = None
                        prev_to = None
                        for (point_from, point_to) in zip(source_pts, target_pts):
                            new_from = bm.verts.new(point_from)
                            new_to = bm.verts.new(point_to)
                            if prev_from:
                                new_face = bm.faces.new([prev_from, prev_to, new_to, new_from])
                            prev_from = new_from
                            prev_to = new_to
                        out_bvht[-1].append(BVHTree.FromBMesh(bm))
                        bm.free()
                    else:
                        out_bvht[-1].append(None)

        return out_bvht


    def before_start(self):
        visible_pockets_ids = self.get_visible_pockets()
        visible_pockets_count = len(visible_pockets_ids)
        self.sewing_id_offset = 0
        self.sorted_hover_sewings = []
        if visible_pockets_count == 0:
            self.pocket_index = -1
        else:
            self.pocket_index = visible_pockets_ids[0]
        self.pocket_sewing_idx = -1

        if self.pocket_index != -1:
            current_pocket = self.garment.pockets[self.pocket_index]
            if current_pocket.pattern_obj:
                self.deselect_all()

        for pocket_id in visible_pockets_ids:
            pocket = self.garment.pockets[pocket_id]
            if pocket.pattern_obj and pocket.target_pattern:
                self.cache_target_points(pocket_id)
        self.pocket_sewings_bvht = self.update_pockets_bvht()


    def modal(self, context, event):
        self.context = context
        self.event = event
        context.area.tag_redraw()

        if self.done:
            return {self.done}

        self.fsm_update()
        if self.allow_nav:
            self.allow_nav = False
            return {'PASS_THROUGH'}

        return {'RUNNING_MODAL'}


    def invoke(self, context, event):
        self.garment_index = context.scene.garment_index
        if context.space_data.type == 'VIEW_3D':
            self.invoke_garment(context, event)
            self.fsm_init()
            self.before_start()
            context.window_manager.modal_handler_add(self)
            return {'RUNNING_MODAL'}
        else:
            self.report({'WARNING'}, "Active space must be a View3d")
            return {'CANCELLED'}
