# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# Copyright (C) 2017 JOSECONSCO
# Created by JOSECONSCO

import bpy
import blf
import time
import os
import gpu
import numpy as np
import bmesh
from gpu_extras.batch import batch_for_shader

import subprocess
from ..utils.helper_functions import get_addon_preferences
from .garments_lib import copy_garment_data
from ..garments_library import garments_lib

shader_2d = gpu.shader.from_builtin('UNIFORM_COLOR')


def draw_background_box(width, height):
    x0 = 5
    y0 = 60

    vertices = (
        (x0, y0), (width, y0),
        (x0, height), (width, height))
    indices = ((0, 1, 2), (2, 1, 3))
    gpu.state.blend_set("ALPHA")
    batch = batch_for_shader(shader_2d, 'TRIS', {"pos": vertices}, indices=indices)
    shader_2d.bind()
    shader_2d.uniform_float("color", (0., 0., 0., 0.3))
    batch.draw(shader_2d)
    gpu.state.blend_set("NONE")

def draw_screen_rect_px(self, context):
    font_id = 0
    draw_background_box(370, 180)

    if bpy.app.version < (4, 0, 0):
        blf.size(font_id, 22, 65)
    else: # for blender 4 4 and up
        blf.size(font_id, 22)

    blf.position(font_id, 25, 155, 0)
    blf.draw(font_id, "Generating Thumbnail for preset")
    if bpy.app.version < (4, 0, 0):
        blf.size(font_id, 16, 60)
    else: # for blender 4 4 and up
        blf.size(font_id, 16)

    blf.position(font_id, 25, 135, 0)
    blf.draw(font_id, "Draw rectangle with LMB in 3D View")
    blf.position(font_id, 25, 115, 0)
    blf.draw(font_id, "Drag corner points to adjust box size")
    blf.position(font_id, 25, 95, 0)
    blf.draw(font_id, "ENTER - confirm and save garment")
    blf.position(font_id, 25, 75, 0)
    blf.draw(font_id, "ESC / RMB - Cancel")

    gpu.state.point_size_set(6)
    if self.state in ['second_point', 'box_finished', 'tweak_box']:
        start_point = self.mouse_start
        end_point = self.mouse_end if self.mouse_end else self.mouse_pos

        coords = [(start_point[0], start_point[1]), (start_point[0], end_point[1]),
                    (end_point[0], end_point[1]), (end_point[0], start_point[1]),
                    (start_point[0], start_point[1])]

        batch = batch_for_shader(shader_2d, 'LINE_STRIP', {"pos": coords})
        shader_2d.bind()
        shader_2d.uniform_float("color", (0.7, 0.7, 0.7, 1))
        batch.draw(shader_2d)

        coords_points = [(start_point[0], start_point[1]),  (end_point[0], end_point[1])]
        batch = batch_for_shader(shader_2d, 'POINTS', {"pos": coords_points})
        shader_2d.bind()
        batch.draw(shader_2d)

        gpu.state.point_size_set(1)


def screen_grab_new(self, context, output_png_path):
    img_name = 'vieport_thumb_oqn'
    low_x = min(self.mouse_start[0], self.mouse_end[0]) #+ context.region.x # to get absolute window co
    low_y = min(self.mouse_start[1], self.mouse_end[1]) #+ context.region.y  # to get absolute window co
    width = max(self.mouse_start[0], self.mouse_end[0]) - low_x  #+ context.region.x # to get absolute window co
    height = max(self.mouse_start[1], self.mouse_end[1]) - low_y # + context.region.y  # to get absolute window co
    img = bpy.data.images.get(img_name)
    if not img:
        img = bpy.data.images.new(img_name , width, height, alpha=True)
    else:
        img.scale(width, height)


    def render_viewport_color(context, rwidth=None, rheight=None, view_matrix=None, projection_matrix=None):
        offscreen = gpu.types.GPUOffScreen(rwidth, rheight)
        color = None
        with offscreen.bind():
            fb = gpu.state.active_framebuffer_get()
            # fb.clear(color=(1.0, 0.0, 0.0, 0.0), depth=1)
            gpu.state.depth_test_set('LESS_EQUAL')
            gpu.state.depth_mask_set(True)
            with gpu.matrix.push_pop():
                gpu.matrix.load_matrix(view_matrix)
                gpu.matrix.load_projection_matrix(projection_matrix)
                offscreen.draw_view3d(
                    context.scene,
                    context.view_layer,
                    context.space_data,
                    context.region,
                    view_matrix,
                    projection_matrix,
                    do_color_management=True
                )
            color = np.array(fb.read_color(low_x, low_y, width, height, 4, 0, 'FLOAT').to_list())
        gpu.state.depth_test_set('NONE')
        offscreen.free()
        return color

    # view_matrix = context.region_data.view_matrix
    # perspective_matrix = context.region_data.perspective_matrix # view projection matrix ?

    view3d = context.space_data
    view_matrix = view3d.region_3d.view_matrix
    perspective_matrix = view3d.region_3d.perspective_matrix # perspective_matrix == window_matrix * view_matrix
    window_matrix = view3d.region_3d.window_matrix # == projection_matrix for scene.camera

    # view_matrix = context.scene.camera.matrix_world.inverted()
    # projection_matrix = context.scene.camera.calc_matrix_camera( # XXX: requires CAM
    #     bpy.context.evaluated_depsgraph_get(),
    #     x=width,
    #     y=height)


    pixelBuffer = render_viewport_color(context, context.region.width, context.region.height, view_matrix, window_matrix)


    # pixels = np.array(pixelBuffer, dtype='f').reshape(width, height, 4)  # RGBA
    # img.pixels = np.power(pixelBuffer.ravel(), 0.454) # we should split Alpha in theory but the effect is cool this way
    img.pixels = pixelBuffer.ravel() # do_color_management=True == get rid of inverse gamma correction
    avg_size = (width + height)/2
    k = 180 / avg_size
    img.scale(int(width * k), int(height * k))
    img.filepath_raw = output_png_path  # blender needs slash at end
    img.save()
    bpy.data.images.remove(img)

    self.fin = True

class GT_OT_ModalLibStore(bpy.types.Operator):
    bl_idname = "garment_lib.save_garment"
    bl_label = "Store garment"
    bl_description = "Save selected scene garment(s) to library"
    bl_options = {"REGISTER"}

    with_screen_grab: bpy.props.BoolProperty(name='With screen grab', description='With screen grab',  default=False)

    file_exits = False
    confirm_ask = True
    #for modal (with_screen_grab) version
    _handle = None
    mouse_start = False
    mouse_end = False
    mouse_pos = []  # we need it to pass as self.mouse_pos to draw handler
    state = 'first_point'

    def modal(self, context, event):
        if self.fin:
            context.area.tag_redraw()
            bpy.types.SpaceView3D.draw_handler_remove(self._handle_scgrab, 'WINDOW')
            garments_lib.FORCE_UPDATE_THUMBS = True
            self.report({'INFO'}, 'Export was successful')
            return {'FINISHED'}

        context.area.tag_redraw()
        if event.type == 'MOUSEMOVE':
            self.mouse_pos = [event.mouse_region_x, event.mouse_region_y]
            if self.state == 'tweak_box':
                x = event.mouse_region_x
                y = event.mouse_region_y
                delta = 33
                if x-delta < self.mouse_start[0] < x + delta and y-delta < self.mouse_start[1] < y + delta:
                    self.mouse_start = [x, y]
                elif x-delta < self.mouse_end[0] < x + delta and y-delta < self.mouse_end[1] < y + delta:
                    self.mouse_end = [x, y]


        if self.state == 'first_point' and event.type == 'LEFTMOUSE' and event.value == 'PRESS':  # PRESS
            self.mouse_start = [event.mouse_region_x, event.mouse_region_y]
            self.state = 'second_point'

        elif self.state == 'second_point' and event.type == 'LEFTMOUSE' and event.value == 'RELEASE':  # finish drawing box, and calculate uv Transformation
            self.mouse_end = [event.mouse_region_x, event.mouse_region_y]
            self.state = 'box_finished'

        elif self.state == 'box_finished' and event.type == 'LEFTMOUSE' and event.value == 'PRESS':  # finish drawing box, and calculate uv Transformation
            self.state = 'tweak_box'

        elif self.state == 'tweak_box' and event.type == 'LEFTMOUSE' and event.value == 'RELEASE':  # finish drawing box, and calculate uv Transformation
            self.state = 'box_finished'

        elif event.type in {'RIGHTMOUSE', 'ESC'}:
            if self._handle:
                bpy.types.SpaceView3D.draw_handler_remove(self._handle, 'WINDOW')
            self.report({'INFO'}, 'CANCELLED')
            return {'CANCELLED'}

        elif event.type in {'NUMPAD_ENTER', 'RET'} and event.value == 'PRESS':
            if self._handle:
                bpy.types.SpaceView3D.draw_handler_remove(self._handle, 'WINDOW')
            context.area.tag_redraw()


            g_prop = context.scene.garment_props  # if multiple garments, then get selected for export from here
            prefs = get_addon_preferences()

            current_folder = os.path.join(prefs.lib_paths[prefs.lib_active_index].directory, g_prop.lib_relative_sub_path)
            garment_ids = [int(i) for i in g_prop.export_garments_li] if len(context.scene.cloth_garment_data) > 1 else [0]
            blend_name = g_prop.export_merged_name if len(garment_ids) > 1 else context.scene.cloth_garment_data[garment_ids[0]].name
            output_blend_file = os.path.join(current_folder, blend_name+'.blend')
            output_png = output_blend_file[:-5] + "png"
            screen_grab_new(self, context, output_png)

            save_garment_to_blend(context, output_blend_file)
            # enum_previews_from_directory_items(self, context)
            # self._handle_scgrab = bpy.types.SpaceView3D.draw_handler_add(screen_grab_new, (self,output_png, context), 'WINDOW', 'PRE_VIEW') # this draws the viewport objects alone (without grid)
            self.report({'INFO'}, 'Export was successful')
            bpy.app.timers.register(delayed_thumbs_update, first_interval=0.5)  # delay for background finish render preview. Then update
            context.area.tag_redraw()
            return {'FINISHED'}


        elif event.type in {'MIDDLEMOUSE', 'WHEELUPMOUSE', 'WHEELDOWNMOUSE'}:
            # allow navigation
            return {'PASS_THROUGH'}

        return {'RUNNING_MODAL'}

    def invoke(self, context, event):
        self.confirm_ask = True
        self.fin = False
        return context.window_manager.invoke_props_dialog(self)


    def run_mod(self, context):
        ''''Run modal save, with scrreen grab'''

        self.startTime = time.time()
        self.fin = False
        self.state = 'first_point'
        self.mouse_pos = [0, 0]
        self.mouse_start = False
        self.mouse_end = False
        args = (self, context)
        self._handle = bpy.types.SpaceView3D.draw_handler_add(draw_screen_rect_px, args, 'WINDOW', 'POST_PIXEL')
        context.window_manager.modal_handler_add(self)
        self.report({'INFO'}, 'Draw box shape 3D View')
        return {'RUNNING_MODAL'}


    def draw(self, context):
        g_props = context.scene.garment_props
        layout = self.layout
        if self.file_exits:
            layout.label(text='File already exist. Pressing Ok will override it!')
        else:
            col = layout.column(align=True)
            if len(context.scene.cloth_garment_data) > 1:
                row = col.row()
                row.prop(g_props, "export_garments_li", expand=True)
            else: #there is at least one, cos of poll
                layout.label(text=f'Export preset: {context.scene.cloth_garment_data[0].name}')
            if len(g_props.export_garments_li) > 1: #if user selects at least two presets for export, show merged name popup
                row = col.row()
                row.prop(g_props, "export_merged_name", text='Name')
                # row.prop(g_props, 'save_with_screen_grab', icon='IMAGE_PLANE', text='')

    def execute(self, context):
        if self.with_screen_grab:  # run modal screen grab
            g_prop = context.scene.garment_props
            prefs = get_addon_preferences()
            garment_ids = [int(i) for i in g_prop.export_garments_li] if len(context.scene.cloth_garment_data) > 1 else [0]
            if len(garment_ids) == 0:  # there is always at least one garment. Bug user may chave not chosen any
                self.report({'ERROR'}, 'Nothing was selected for export! Cancelling')
                return {'CANCELLED'}

            patterns = 0
            for g_id in garment_ids:
                garment = context.scene.cloth_garment_data[g_id]
                for pattern in garment.sewing_patterns:
                    if pattern.pattern_obj:
                        patterns += 1

            if patterns == 0:
                self.report({'ERROR'}, 'Selected garment has no sewing patterns! Cancelling')
                return {'CANCELLED'}

            g_prop = context.scene.garment_props
            prefs = get_addon_preferences()
            garment_ids = [int(i) for i in g_prop.export_garments_li] if len(context.scene.cloth_garment_data) > 1 else [0]

            current_folder = os.path.join(prefs.lib_paths[prefs.lib_active_index].directory, g_prop.lib_relative_sub_path)
            blend_name = g_prop.export_merged_name if len(garment_ids) > 1 else context.scene.cloth_garment_data[0].name
            output_blend_file = os.path.join(current_folder, blend_name+'.blend')

            self.file_exits = os.path.isfile(output_blend_file)
            if self.file_exits and self.confirm_ask: #ask just once if override
                self.confirm_ask = False
                return context.window_manager.invoke_props_dialog(self)
            self.run_mod(context)
            return {'RUNNING_MODAL'}

        output_blend_file = save_garment_to_blend(context)
        render_preview_background(context, output_blend_file)
        # enum_previews_from_directory_items(self, context)
        bpy.app.timers.register(delayed_thumbs_update, first_interval=3)  # delay for background finish render preview. Then update

        self.report({'INFO'}, 'Export was successful')
        return {"FINISHED"}

def render_preview_background(context, blend_file):
    command = [bpy.app.binary_path,
               blend_file,
               '-b',
               '--python',
               os.path.join(os.path.dirname(os.path.realpath(__file__)), 'generate_thumb.py')]
    subprocess.Popen(command)


def delayed_thumbs_update():
    garments_lib.FORCE_UPDATE_THUMBS = True


def save_garment_to_blend(context, output_blend_file):
    g_prop = context.scene.garment_props  # if multiple garments, then get selected for export from here
    garment_ids = [int(i) for i in g_prop.export_garments_li] if len(context.scene.cloth_garment_data) > 1 else [0]
    map_new_old_garment_id = {i: garment_ids.index(i) for i in range(len(context.scene.cloth_garment_data)) if i in garment_ids}

    for obj in context.selected_objects:
        obj.select_set(False)

    exp_scene = bpy.data.scenes.new('garment_export')
    backup_cloth_names = {}
    for g_id in garment_ids:
        copy_garment_data(context.scene, exp_scene, g_id)
        garment = context.scene.cloth_garment_data[g_id]
        for pattern in garment.sewing_patterns[:]+garment.pockets[:]:
            if pattern.pattern_obj:
                exp_scene.collection.objects.link(pattern.pattern_obj)

        #copy obj with cloth sim if exist
        if garment.target_cloth_object:
            cloth_obj = garment.target_cloth_object
            if cloth_obj.type != 'MESH':
                continue

            mesh = bpy.data.meshes.new("export_meh")  # add a new mesh
            cloth_copy_exp = bpy.data.objects.new("mesh_cloth_export", mesh)  # add a new object using the mesh
            exp_scene.collection.objects.link(cloth_copy_exp)
            exp_scene.collection.objects.link(cloth_obj)
            bm = bmesh.new()
            v1 = bm.verts.new((-0.1, 0, 0))
            v2 = bm.verts.new((0.1, 0, 0))
            bm.edges.new([v1, v2])
            bm.to_mesh(mesh)
            bm.free()
            for vg in cloth_obj.vertex_groups:
                cloth_copy_exp.vertex_groups.new(name=vg.name)
            back_scene = context.window.scene
            context.window.scene = exp_scene
            context.view_layer.objects.active = cloth_obj
            cloth_obj.select_set(True)
            cloth_copy_exp.select_set(True)
            context.window.scene = exp_scene
            bpy.ops.object.editmode_toggle()
            bpy.ops.object.editmode_toggle()

            bpy.ops.object.make_links_data(type='MODIFIERS')
            exp_scene.collection.objects.unlink(cloth_obj)
            cloth_copy_exp.hide_set(True)
            context.window.scene = back_scene
            cloth_copy_exp.name = garment.target_cloth_object.name
            new_g_idx = map_new_old_garment_id[g_id]
            exp_scene.cloth_garment_data[new_g_idx].target_cloth_object = cloth_copy_exp
            backup_cloth_names[cloth_obj.name] = garment.target_cloth_object.name
            # for mod in cloth_copy_exp.modifiers:
            #     if mod.type == 'CLOTH':
            #         mod.show_viewport = False

    data_blocks = set([exp_scene])
    bpy.data.libraries.write(output_blend_file, data_blocks, fake_user=True)
    bpy.data.scenes.remove(exp_scene)
    for current_name, original_name in backup_cloth_names.items():  # restore original names to garment
        bpy.data.objects[current_name].name = original_name
    return output_blend_file
