# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# Copyright (C) 2017 JOSECONSCO
# Created by JOSECONSCO

from ..utils.helper_functions import draw_text_line
from ..utils.fsm_oper import ModalCommon, CookieCutter_FSM
from mathutils.bvhtree import BVHTree
from mathutils import Vector
from gpu_extras.batch import batch_for_shader
import time
import gpu
import gpu
import bpy
from copy import deepcopy
from .spline_base import *
import logging

# logging.basicConfig(format='%(lineno)d - %(message)s Dupa', level=logging.DEBUG)  # wtf wont work
logger = logging.getLogger()
logger.disabled = True
# A helper function to filter for values in the [0,1] interval:

def add_sewing(pattern, from_spl_id, from_seg_id, to_spl_id, to_seg_id, flip=False):
    for garment in bpy.context.scene.cloth_garment_data:
        existing_patterns = [g_mesh.pattern_obj.name for g_mesh in garment.sewing_patterns if g_mesh.pattern_obj]
        if pattern.name in existing_patterns:  # find garments  that are using active_obj
            garment.garment_sewings.add()
            garment.garment_sewings[-1].source_obj = pattern
            garment.garment_sewings[-1].target_obj = pattern
            garment.garment_sewings[-1].from_spline_idx = from_spl_id
            garment.garment_sewings[-1].from_segment_idx = from_seg_id
            garment.garment_sewings[-1].to_spline_idx = to_spl_id 
            garment.garment_sewings[-1].to_segment_idx = to_seg_id
            garment.garment_sewings[-1].flip = flip #!


class GTOOL_OT_RayCastKnife(bpy.types.Operator, ModalCommon, CookieCutter_FSM):
    bl_idname = "curve.pattern_cut"
    bl_label = "Pattern Knife Tool"
    bl_description = "Pattern Knife Tool"
    bl_options = {'REGISTER', 'UNDO'}

    garment_index: bpy.props.IntProperty()
    pocket_index: bpy.props.IntProperty(default=0)

    def draw_text(self, cla, context):
        draw_text_lines = []
        draw_text_lines.append(["Click to cut - LMB", 'H1'])
        # draw_text_lines.append([f"Is: {self.connect_pts_li[0].placement}", 'H1'])
        # if self._state == 'main':
        # draw_text_lines.append([f"Click pos: {self.prev_click_pos}", 'H1'])
        draw_text_line(draw_text_lines)


    def draw_px(self, cla, context):
        if self._state in ['main', 'knife_cut']:  # hover highlight
            shader = self.shader
            gpu.state.blend_set("ALPHA")
            gpu.state.point_size_set(6)
            gpu.state.line_width_set(2.0)
            gpu.state.depth_test_set("LESS")

            if self.gl_cut_pts_world:
                batch = batch_for_shader(shader, 'LINE_STRIP', {"pos": self.gl_cut_pts_world})
                shader.bind()  # draw currently defined sewing
                shader.uniform_float("color", (0.9, 0.9, 0.9, 0.7))
                batch.draw(shader)

                batch = batch_for_shader(shader, "POINTS", {"pos": self.gl_cut_pts_world})
                shader.bind()
                shader.uniform_float("color", (0.9, 0.9, 0.9, 0.7))
                batch.draw(shader)

            batch = batch_for_shader(shader, "POINTS", {"pos": [self.gl_hover_co]})
            shader.bind()
            shader.uniform_float("color", (0.2, 0.9, 0.2, 0.8))
            batch.draw(shader)

            #restore opengl defaults
            gpu.state.depth_test_set("NONE")
            gpu.state.line_width_set(1)
            gpu.state.point_size_set(1)
            gpu.state.blend_set("NONE")


    def connect_two_pts(self, cut_pts):
        # spl_slice.extend([{'spl_id': spl_id, 'p_id': p_id} for p_id, p in enumerate(spl.bezier_points) if p.select_control_point])
        b_splines = self.b_splines
        delta = 0.001
        curve_obj = b_splines.curve_obj
        point_1 = cut_pts[0]
        point_2 = cut_pts[1]
        logging.debug(cut_pts[0].__dict__)
        logging.debug(cut_pts[1].__dict__)
        # import pprint
        # pprint.pprint(cut_pts[0].__dict__)
        # pprint.pprint(cut_pts[1].__dict__)
        # slice_type = 'SPLIT_SLICE' - when we cut outline in half
        if point_1.spl_id == point_2.spl_id:  # we split outline in 2 parts it seems !what about holes ?
            sliced_spline = b_splines[point_1.spl_id]
            spl_ab, spl_ba = sliced_spline.split_by_two_ids((point_1.p_id, point_2.p_id))

            spl_ab.points[-1].sewing.spl_id = -1
            spl_ba.points[-1].sewing.spl_id = -1

            lerp_fac_ab = [0, 0]  # do not blend to  dir_fix_prev for spl ab
            lerp_fac_ba = [0, 0]  # do not blend to  dir_fix_prev for spl ba
            #* use average (curr_stroke_dir + prev_stroke_dir)/2 if connecting old inside point
            # if point_1.placement == 'INSIDE' and sliced_spline.length != 1:  # is first point was 'inside' point? (after connectin holes its len > 1)
            #     prev_dir = (self.click_pos[-2] - self.click_pos[-3]).normalized()
            #     avg_dir = (prev_dir + forward_dir).normalized()  # average prev dir and current dir
            #     dir_fix_prev = avg_dir.yxz  # fix dir for inside cuts
            #     dir_fix_prev.x *= -1

            #     first_dir = (self.click_pos[1] - self.click_pos[0]).normalized()
            #     avg_dir = (first_dir + forward_dir).normalized()  # average prev dir and current dir
            #     dir_fix_first = avg_dir.yxz  # fix dir for inside cuts
            #     dir_fix_first.x *= -1

            #     p1_id_is_smaller = True if point_1.p_id < point_2.p_id else False
            #     lerp_fac_ab = [1, 0] if p1_id_is_smaller else [0, 1] # for spl_ab -> blend to 
            #     lerp_fac_ba = [0, 1] if p1_id_is_smaller else [1, 0] # for spl_ab -> blend to 
            # else:
            #     dir_fix_prev = perpendicular_dir
            #     dir_fix_first = perpendicular_dir

            forward_dir = (spl_ab[-1].co-spl_ab[0].co).normalized()  # cos they are sorted, split_by_two_ids so its consistient
            perpendicular_dir = forward_dir.yxz  # rot right
            perpendicular_dir.x *= -1
            
            spl_ab[0].handle_right_type = 'FREE'  
            spl_ab[0].handle_left_type = 'VECTOR'  # sharp
            spl_ab[-1].handle_right_type = 'VECTOR' # sharp
            spl_ab[-1].handle_left_type = 'FREE'  

            spl_ba[0].handle_right_type = 'FREE'
            spl_ba[0].handle_left_type = 'VECTOR'# sharp
            spl_ba[-1].handle_right_type = 'VECTOR'  # sharp
            spl_ba[-1].handle_left_type = 'FREE'

            ab_is_inside = spl_ba.is_point_inside_winding(spl_ab.get_center())
            ba_is_inside = spl_ab.is_point_inside_winding(spl_ba.get_center())
            are_overlapping = True if ab_is_inside or ba_is_inside else False # 8 or (0) type
            ab_dir = spl_ab.get_signed_area()
            ba_dir = spl_ba.get_signed_area()

            # if are_overlapping: #falculate n+1 and n-1 forward dirs, and perpendicular dirs
            ab_forward = ((spl_ab[0].co - spl_ab[1].co).normalized() + forward_dir)/2
            ab_f_perp = ab_forward.yxz
            ab_f_perp.x *= -1
            ab_backward = ((spl_ab[-2].co - spl_ab[-1].co).normalized() + forward_dir)/2
            ab_b_perp = ab_backward.yxz
            ab_b_perp.x *= -1
            ba_forward = ((spl_ba[0].co - spl_ba[1].co).normalized() - forward_dir)/2 #cos forward is -forw on ba
            ba_f_perp = ba_forward.yxz
            ba_f_perp.x *= -1
            ba_backward = ((spl_ba[-2].co - spl_ba[-1].co).normalized() - forward_dir)/2  # cos forward is -forw on ba
            ba_b_perp = ba_backward.yxz
            ba_b_perp.x *= -1
            # else:
            #     ab_f_perp = ab_b_perp = perpendicular_dir
            #     ba_f_perp = ba_b_perp = -1 * perpendicular_dir

            ab_is_outside = True if abs(spl_ab.get_signed_area()) > abs(spl_ba.get_signed_area()) else False
            flip = 1 if not are_overlapping or (are_overlapping and not ab_is_outside) else -1  # 8 type of split
            if ab_dir > 0:
                spl_ab[0].co += flip * delta * ab_f_perp
                spl_ab[-1].co += flip * delta * ab_b_perp
                spl_ba[0].co += flip * delta * ba_f_perp  # automatically ba_dir > 0: (if not are_overlapping)
                spl_ba[-1].co += flip * delta * ba_b_perp
            else:
                spl_ab[0].co -= flip * delta * ab_f_perp
                spl_ab[-1].co -= flip * delta * ab_b_perp
                spl_ba[0].co -= flip * delta * ba_f_perp  # automatically ba_dir < 0: (if not are_overlapping)
                spl_ba[-1].co -= flip * delta * ba_b_perp
                
            #! first write to new spline (-1), so that second write_to_blender_spl wont override first wtbs sewings
            last_id = len(curve_obj.data.splines)
            spl_ba.write_to_blender_spl(curve_obj, spl_id=last_id)  # will crate new one, cos id = len
            spl_ab.write_to_blender_spl(curve_obj, spl_id=point_1.spl_id)
            add_sewing(curve_obj, from_spl_id=last_id, from_seg_id=spl_ba.length-1, to_spl_id=point_1.spl_id, to_seg_id=spl_ab.length-1, flip=True)

            b_splines.append(spl_ba)
            b_splines[point_1.spl_id] = spl_ab
            return

        # slice_type = 'CONNECT_SLICE' - connect 2 holes
        elif point_1.spl_id != point_2.spl_id:  # we connect 2 holes in 1 it seems
            hole1_spl_idx, hole1_cut_idx = (point_1.spl_id, point_1.p_id)
            hole2_spl_idx, hole2_cut_idx = (point_2.spl_id, point_2.p_id)

            hole1_spl = b_splines[hole1_spl_idx]
            hole2_spl = b_splines[hole2_spl_idx]

            #* add pt to hole 1
            h1_len = hole1_spl.length
            if h1_len > 1:
                hole1_spl.insert(hole1_cut_idx, [hole1_spl[hole1_cut_idx]])  # we split cut point on hole_bezier spline on 2
                hole1_spl[hole1_cut_idx].sewing.spl_id = -1  # unlink one of split points #? maybe bridge this...
                hole1_spl[hole1_cut_idx].handle_left_type = 'FREE'
                hole1_spl[hole1_cut_idx].handle_right_type = 'VECTOR'  # sharp
                hole1_spl[hole1_cut_idx+1].handle_left_type = 'VECTOR'  # sharp
                hole1_spl[hole1_cut_idx+1].handle_right_type = 'FREE'  
            else:
                hole1_spl[hole1_cut_idx].handle_left_type = 'VECTOR'
                hole1_spl[hole1_cut_idx].handle_right_type = 'VECTOR'

            #* add pt to hole 2
            h2_len = hole2_spl.length
            if h2_len > 1:
                hole2_spl.insert(hole2_cut_idx, [hole2_spl[hole2_cut_idx]])  # we split cut point on hole_bezier spline on 2
                hole2_spl.rotate(-1*hole2_cut_idx-1)  # move cut to begining of slice, minus one for vert clone

                hole2_spl[-1].handle_left_type = 'FREE'
                hole2_spl[-1].handle_right_type = 'VECTOR'  # sharp
                hole2_spl[0].handle_left_type = 'VECTOR'  # sharp
                hole2_spl[0].handle_right_type = 'FREE'
            else:
                hole2_spl[0].handle_right_type = 'VECTOR'
                hole2_spl[0].handle_left_type = 'VECTOR'

            h1_is_inside = hole2_spl.is_point_inside_winding(hole1_spl[hole1_cut_idx].co)
            h2_is_inside = hole1_spl.is_point_inside_winding(hole2_spl[0].co)
            flip = -1 if h1_is_inside or h2_is_inside else 1

            h1_sig_area = hole1_spl.get_signed_area()
            h2_sig_area = hole2_spl.get_signed_area()
            #* merge separate splines.
            if flip * h1_sig_area * h2_sig_area > 0:  # for cutting hole in outline, both splines have to have opposite direction (sig_area)
                hole2_spl[0].sewing.spl_id = -1  # unlink one of split points
                hole1_spl.insert(hole1_cut_idx+1, hole2_spl)  # in outline split place we put the hole spline, +1 to offset vert clone
            else:
                hole2_spl[-1].sewing.spl_id = -1  # unlink one of split points
                hole1_spl.insert(hole1_cut_idx+1, hole2_spl.reversed())  # in outline split place we put the hole spline #!!!

            #* offset connected verts slightly away from  cut line to avoid filling errors
            spl_are_donut_type = True if flip == 1 else False  # donuut type curves or 8
            h1_is_cw = h1_sig_area > 0
            h2_is_cw = h2_sig_area > 0

            if spl_are_donut_type:
                dir_flip = 1 if not h1_is_cw else -1
            else:
                if h1_is_inside:
                    dir_flip = 1 if not h1_is_cw else -1
                else:  # h2 must be inside h1 loop
                    dir_flip = 1 if h1_is_cw else -1

            forward_dir = (point_2.click_pos - point_1.click_pos).normalized()

            perpendicular_dir = forward_dir.yxz
            perpendicular_dir.x *= -1

            #* use average (curr_stroke_dir + prev_stroke_dir)/2 if connecting old inside point
            if point_1.placement == 'INSIDE' and h1_len != 1:  # is first point was 'inside' point? (after connectin holes its len > 1)
                prev_dir = (self.click_pos[-2] - self.click_pos[-3]).normalized()
                avg_dir = (prev_dir + forward_dir).normalized()  # average prev dir and current dir
                perpendicular_dir_h1 = avg_dir.yxz  # fix dir for inside cuts
                perpendicular_dir_h1.x *= -1
            else:
                perpendicular_dir_h1 = perpendicular_dir

            if h1_len>1: #h1 fix - otherwise - seg count
                hole1_spl[hole1_cut_idx].co -= dir_flip * delta * perpendicular_dir_h1  # fix hole1[hole1_cut_idx] point
                hole1_spl[hole1_cut_idx + h2_len+1].co += dir_flip * delta * perpendicular_dir_h1  # fix right old hole1[hole1_cut_idx+1] point

            if h2_len>1: #h2 fix  - otherwise - seg count
                hole1_spl[hole1_cut_idx + 1].co -= dir_flip * delta * perpendicular_dir  # fix left h2[0 point
                hole1_spl[hole1_cut_idx + h2_len].co += dir_flip * delta * perpendicular_dir  # fix right h2[0] point

            merged_spl_id = min(hole1_spl_idx, hole2_spl_idx)
            hole1_spl.write_to_blender_spl(curve_obj, spl_id=merged_spl_id)

            from_seg_id = hole1_cut_idx
            to_seg_id = hole1_cut_idx + h2_len + 1 if h2_len > 1 else hole1_cut_idx + h2_len
            add_sewing(curve_obj, from_spl_id=merged_spl_id, from_seg_id=from_seg_id, to_spl_id=merged_spl_id, to_seg_id=to_seg_id, flip=True)

            #* if we were connecting to single point spline (inside pattern point) return it for next step cut
            ret_point = CutPointInfo(spl_id=merged_spl_id, p_id=hole1_cut_idx+1, placement='INSIDE', click_pos=hole1_spl[hole1_cut_idx + 1].co)  # return connected point if single.

            #* after connecting 2 splines remove last one, and its sewings
            del_spl_id = max(hole1_spl_idx, hole2_spl_idx)
            del(b_splines[del_spl_id])
            curve_obj.data.splines.remove(curve_obj.data.splines[del_spl_id])
            sewing_rem_spline_id(self.context, curve_obj, del_spl_id, remap_removed_spline=False)
            return ret_point
    
    def calc_mouse_hover(self, only_hover = True):
        ''' Usually just for GL drawing (sets self.gl_hover_co), but on only_hover = False:
        returns - click_pos, closest_ctrl_pt'''
        context = self.context
        scene = context.scene
        region = context.region
        rv3d = context.region_data
        coord = self.event.mouse_region_x, self.event.mouse_region_y

        # get the ray from the viewport and mouse
        view_vector = view3d_utils.region_2d_to_vector_3d(region, rv3d, coord)
        ray_origin = view3d_utils.region_2d_to_origin_3d(region, rv3d, coord)

        obj = context.active_object

        ray_target = ray_origin + view_vector

        plane_co = Vector((0, 0, 0))
        plane_no = Vector((0, 0, 1))

        matrix_inv = obj.matrix_world.inverted()
        ray_origin_obj = matrix_inv @ ray_origin
        ray_target_obj = matrix_inv @ ray_target
        # ray_direction_obj = ray_target_obj - ray_origin_obj

        click_pos = mathutils.geometry.intersect_line_plane(ray_origin_obj, ray_target_obj, plane_co, plane_no, True)
        self.gl_hover_co = click_pos

        #* did we click close to control point?
        closest_ctrl_pt, closets_dist, best_ctrl_pt_co = self.b_splines.get_closest_ctrl_point(click_pos)
        best_ctrl_pt_2d = view3d_utils.location_3d_to_region_2d(region, rv3d, obj.matrix_world @ best_ctrl_pt_co, default=None)
        # self.best_ctrl_pt_2d = best_ctrl_pt_2d.copy()

        #* did we click near spline?
        closest_spl_id, closest_spl_point, closest_tn, closest_dist = self.b_splines.closest_spl_point(click_pos)
        closest_spl_pt_2d = view3d_utils.location_3d_to_region_2d(region, rv3d, obj.matrix_world @ closest_spl_point, default=None)
        # self.closest_spl_pt_2d = closest_spl_pt_2d.copy()

        click_projected_2d = view3d_utils.location_3d_to_region_2d(region, rv3d, obj.matrix_world @ click_pos, default=None)
        # self.click_projected_2d = click_projected_2d.copy()

        if (click_projected_2d - best_ctrl_pt_2d).length < 15:  # close to control point in screen space
            self.gl_hover_co = best_ctrl_pt_co
            click_pos = best_ctrl_pt_co
        else:
            closest_ctrl_pt = None  # cut not from existing ctrl point
            if (click_projected_2d - closest_spl_pt_2d).length < 15:  # close to control point in screen space
                click_pos = closest_spl_point
                self.gl_hover_co = closest_spl_point

                #* add new point in here, as:  closest_ctrl_pt 
                if not only_hover:  # perform actual slice, and set is as closest_ctrl_pt
                    new_pt_id = self.b_splines[closest_spl_id].split_seg_t(closest_tn)  # done: we should offset connect_pts_li, pts_ids after new_pt_id...
                    self.b_splines[closest_spl_id].write_to_blender_spl(self.obj, spl_id=closest_spl_id)  # add new
                    if self.connect_pts_li: #if slice list have some pts, offset them +1
                        for old_pt in self.connect_pts_li:  # for new points
                            if old_pt.spl_id == closest_spl_id and old_pt.p_id >= new_pt_id: #old pt is behind new_pt, so id+=1
                                old_pt.p_id += 1  # cut point was added before prev_pt, so offset +1

                    closest_ctrl_pt = (closest_spl_id, new_pt_id)

        self.gl_hover_co = obj.matrix_world @ self.gl_hover_co
        return click_pos, closest_ctrl_pt 


    def fill_slices_list(self, first_cut = False):
        '''fills self.connect_pts_li for slicing '''
        click_pos, closest_ctrl_pt = self.calc_mouse_hover(only_hover=False)

        if not first_cut and not closest_ctrl_pt:  # get actual knife cut by intersection
            prev_cut_pts_exist = len(self.connect_pts_li) > 0
            # print('splines_line_intersect. Not first cut')
            self.connect_pts_li.extend(self.b_splines.splines_line_intersect(self.click_pos[-1], click_pos))

            #* if there were exising point before intersection above, check if we have to offset them +=1
            if prev_cut_pts_exist: 
                old_pt = self.connect_pts_li[0]
                for new_pt in self.connect_pts_li[1:]: #for new points
                    if new_pt.spl_id == old_pt.spl_id and old_pt.p_id >= new_pt.p_id:  # old pt is behind new_pt, so id+=1
                        old_pt.p_id += 1  # cut point was added before old_pt, so offset +1

        #*  was last click pos close to control point or inside pattern or outiside?
        if closest_ctrl_pt:  # close to control point in screen space
            # print(f'clost to ctrl pt: {closest_ctrl_pt[0]}:{closest_ctrl_pt[1]}')
            self.connect_pts_li.append(CutPointInfo(spl_id=closest_ctrl_pt[0], p_id=closest_ctrl_pt[1] , placement='BORDER', click_pos=click_pos))
            self.clear_pts_li = True
        else:
            #* If click pos inside sewing pattern - add spline with this single point
            current_click_inside = self.b_splines.is_point_inside_pattern(click_pos.to_2d())
            if current_click_inside:  # seems faster
                print('Clicked inside!')
                new_bp = BezierPoint(click_pos, handle_left_type='VECTOR', handle_right_type='VECTOR', handle_left=click_pos, handle_right=click_pos)
                self.b_splines.append(BezierSpline([new_bp]))
                self.b_splines[-1].write_to_blender_spl(self.obj, spl_id=self.b_splines.length)  # add new
                self.connect_pts_li.append(CutPointInfo(spl_id=self.b_splines.length-1, p_id=0, placement='INSIDE', click_pos=click_pos))  # add new single point for spl_connect
                self.clear_pts_li = False
            else:
                self.clear_pts_li = True
                #self.connect_pts_li.append(CutPointInfo(spl_id=self.b_splines.length-1, p_id=0, placement='OUTSIDE', click_pos=click_pos))  # add new single point for spl_connect

        self.click_pos.append(click_pos)
        self.gl_cut_pts_world.append(self.obj.matrix_world @ click_pos)
        # return hit_bpoint


    @CookieCutter_FSM.FSM_State('main')
    def modal_main(self):
        if self.event.type == 'LEFTMOUSE' and self.event.value == 'PRESS':
            self.fill_slices_list( first_cut=True) #did we hit close to any spline ctrl point?
            self.start_time = time.time()
            return 'knife_cut'

        elif self.event.type == 'MIDDLEMOUSE':
            self.allow_nav = True

        elif self.event.type == 'MOUSEMOVE': 
            self.calc_mouse_hover(only_hover=True)  # sets: self.gl_hover_co, for drawing only

        elif self.event.type in {'RIGHTMOUSE', 'ESC'} and self.event.value == 'PRESS':
            self.end_cancel()
            return

        elif self.event.type == 'RET' and self.event.value == 'PRESS':
            self.end_commit()
            return 'main'


    @CookieCutter_FSM.FSM_State('knife_cut')
    def knife_cut(self):
        if self.event.type == 'LEFTMOUSE' and self.event.value == 'PRESS' and (time.time() - self.start_time) > 0.2:
            self.fill_slices_list( first_cut=False) 
            #* Loop throught connection points, and connect each pair (i_curr, j_next)
            if len(self.connect_pts_li) > 1:
                prev_slice_knot = self.connect_pts_li[0]
                for spl_sl in self.connect_pts_li[1:]:
                    connected_to_single_point = self.connect_two_pts([prev_slice_knot, spl_sl])
                    prev_slice_knot = spl_sl
                self.connect_pts_li = [] if self.clear_pts_li or not connected_to_single_point else [connected_to_single_point]  # init next cut with last point
            self.start_time = time.time()
            return

        elif self.event.type == 'MOUSEMOVE':
            self.calc_mouse_hover(only_hover=True)  # sets: self.gl_hover_co, for drawing only

        elif self.event.type == 'MIDDLEMOUSE':
            self.allow_nav = True

        elif self.event.type in {'RIGHTMOUSE', 'ESC'} and self.event.value == 'PRESS':
            self.end_cancel()
            return

        elif self.event.type == 'RET' and self.event.value == 'PRESS':
            self.end_commit()
            return 'main'

    def before_start(self):
        self.gl_cut_pts_world = [] #for bgl drawing
        self.click_pos = [] #
        self.gl_hover_co = Vector((0,0,0))
        self.obj = self.context.active_object
        self.connect_pts_li = []  # spl_connect = [{'spl_id': spl_id, 'p_id': p_id}]
        self.b_splines = BezierSplines(self.obj)
        self.inside = False

    def modal(self, context, event):
        self.context = context
        self.event = event
        context.area.tag_redraw()

        if self.done:
            return {self.done}

        self.fsm_update()
        if self.allow_nav:
            self.allow_nav = False
            return {'PASS_THROUGH'}

        return {'RUNNING_MODAL'}

    def invoke(self, context, event):
        if context.space_data.type == 'VIEW_3D':
            # self.invoke_garment(context, event)
            self.invoke_basic(context, event)
            self.fsm_init()
            self.before_start()
            context.window_manager.modal_handler_add(self)
            return {'RUNNING_MODAL'}
        else:
            self.report({'WARNING'}, "Active space must be a View3d")
            return {'CANCELLED'}
